from unittest import TestCase

from tmtccmd.runner import initialize_tmtc_commander
from tmtccmd.tm import Service5TM
from tmtccmd.pus.service_1_verification import Service1TMExtended
from tmtccmd.tm.service_5_event import Srv5Subservices
from spacepackets.ccsds.time import CdsShortTimestamp
from tmtccmd.pus.service_17_test import pack_service_17_ping_command
from tmtccmd.utility.tmtc_printer import TmTcPrinter, DisplayMode
from tmtccmd.utility.logger import get_console_logger, set_tmtc_console_logger
from tmtccmd.config.globals import update_global, CoreGlobalIds

from tests.hook_obj_mock import create_hook_mock_with_srv_handlers


class TestPrinter(TestCase):
    def setUp(self):
        self.tmtc_printer = TmTcPrinter()
        self.logger = get_console_logger()
        set_tmtc_console_logger()

    def test_print_functions(self):
        self.assertTrue(self.tmtc_printer.get_display_mode() == DisplayMode.LONG)
        self.tmtc_printer.set_display_mode(DisplayMode.SHORT)
        self.assertTrue(self.tmtc_printer.get_display_mode() == DisplayMode.SHORT)
        self.tmtc_printer.set_display_mode(DisplayMode.LONG)

        service_1_tm = Service1TMExtended(
            subservice=1, time=CdsShortTimestamp.init_from_current_time()
        )
        service_1_packed = service_1_tm.pack()
        self.tmtc_printer.print_telemetry(packet_if=service_1_tm, info_if=service_1_tm)
        # Should not crash and emit warning
        self.tmtc_printer.print_telemetry(packet_if=None, info_if=None)

        self.tmtc_printer.set_display_mode(DisplayMode.SHORT)
        self.tmtc_printer.print_telemetry(packet_if=service_1_tm, info_if=service_1_tm)
        service_1_tm = Service1TMExtended(
            subservice=2, time=CdsShortTimestamp.init_from_current_time()
        )
        service_1_packed = service_1_tm.pack()
        self.tmtc_printer.print_telemetry(
            packet_if=service_1_tm, info_if=service_1_tm, print_raw_tm=True
        )

        self.tmtc_printer.set_display_mode(DisplayMode.LONG)
        service_5_tm = Service5TM(
            subservice=Srv5Subservices.INFO_EVENT,
            object_id=bytearray([0x01, 0x02, 0x03, 0x04]),
            event_id=22,
            param_1=32,
            param_2=82452,
            time=CdsShortTimestamp.init_from_current_time(),
        )
        hook_base = create_hook_mock_with_srv_handlers()
        initialize_tmtc_commander(hook_object=hook_base)

        service_5_packed = service_5_tm.pack()
        self.tmtc_printer.print_telemetry(packet_if=service_5_tm, info_if=service_5_tm)

        hook_base.handle_service_5_event.assert_called_with(
            object_id=bytes([0x01, 0x02, 0x03, 0x04]),
            event_id=22,
            param_1=32,
            param_2=82452,
        )

        service_17_command = pack_service_17_ping_command(ssc=0, apid=42)
        self.tmtc_printer.print_telecommand(
            tc_packet_obj=service_17_command, tc_packet_raw=service_17_command.pack()
        )
        self.tmtc_printer.set_display_mode(DisplayMode.SHORT)
        self.tmtc_printer.print_telecommand(
            tc_packet_obj=service_17_command, tc_packet_raw=service_17_command.pack()
        )
        self.tmtc_printer.set_display_mode(DisplayMode.LONG)

        self.tmtc_printer.clear_file_buffer()

    def tearDown(self) -> None:
        """Reset the hook object"""
        update_global(CoreGlobalIds.TMTC_HOOK, None)



from typing import Dict

from tmtccmd.utility.logger import get_console_logger

LOGGER = get_console_logger()
__OBJECT_ID_DICT = dict()


def insert_object_id(object_id: bytes, object_id_info: list):
    __OBJECT_ID_DICT[object_id] = object_id_info


def insert_object_ids(object_id_dict: Dict[bytes, list]):
    if object_id_dict is not None:
        __OBJECT_ID_DICT.update(object_id_dict)


def get_object_id_info(object_id: bytes):
    return __OBJECT_ID_DICT.get(object_id)



import enum
from typing import Deque, Tuple, List, Union
from spacepackets.ecss.tm import PusTelemetry
from tmtccmd.tm.base import PusTmInfoInterface, PusTmInterface

TelemetryListT = List[bytearray]
TelemetryQueueT = Deque[bytearray]

PusTmQueue = Deque[PusTelemetry]
PusTmTupleT = Tuple[bytearray, PusTelemetry]

PusTmListT = List[PusTelemetry]
PusTmQueueT = Deque[PusTmListT]
PusIFListT = List[Union[PusTmInfoInterface, PusTmInterface]]
PusIFQueueT = Deque[PusIFListT]

PusTmListT = List[PusTelemetry]
PusTmObjQeue = Deque[PusTelemetry]
PusTmTupleQueueT = Deque[PusTmTupleT]


class TmTypes(enum.Enum):
    NONE = enum.auto()
    CCSDS_SPACE_PACKETS = enum.auto()



"""
@file       tmtcc_tm_listener.py
@date       01.11.2019
@brief      Separate class to listen to telecommands.
@author     R. Mueller
"""
import sys
import time
import threading
from collections import deque
from typing import Dict, List, Tuple
from enum import Enum

from spacepackets.ccsds.spacepacket import get_apid_from_raw_space_packet

from tmtccmd.tm.definitions import TelemetryQueueT, TelemetryListT, TmTypes
from tmtccmd.utility.logger import get_console_logger
from tmtccmd.com_if.com_interface_base import CommunicationInterface
from tmtccmd.utility.conf_util import acquire_timeout

LOGGER = get_console_logger()

INVALID_APID = -2
UNKNOWN_TARGET_ID = -1
QueueDictT = Dict[int, Tuple[TelemetryQueueT, int]]
QueueListT = List[Tuple[int, TelemetryQueueT]]


class TmListener:
    """Performs all TM listening operations.
    This listener to have a permanent means to receive data. A background thread is used
    to poll data with the provided communication interface. Dedicated sender and receiver object
    or any other software component can get the received packets from the internal deque container.
    """

    MODE_OPERATION_TIMEOUT = 300
    DEFAULT_UNKNOWN_QUEUE_MAX_LEN = 50
    QUEUE_DICT_QUEUE_IDX = 0
    QUEUE_DICT_MAX_LEN_IDX = 1

    DEFAULT_LOCK_TIMEOUT = 0.5

    class ListenerModes(Enum):
        MANUAL = (1,)
        LISTENER = (2,)
        SEQUENCE = (3,)

    def __init__(
        self,
        com_if: CommunicationInterface,
        tm_timeout: float,
        tc_timeout_factor: float,
        tm_type: TmTypes = TmTypes.CCSDS_SPACE_PACKETS,
    ):
        """Initiate a TM listener.
        :param com_if:              Type of communication interface, e.g. a serial or ethernet interface
        :param tm_timeout:          Timeout for the TM reception
        :param tc_timeout_factor:
        :param tm_type:             Telemetry type. Default to CCSDS space packets for now
        """
        self.__tm_timeout = tm_timeout
        self.tc_timeout_factor = tc_timeout_factor
        self.__com_if = com_if
        # TM Listener operations can be suspended by setting this flag
        self.event_listener_active = threading.Event()
        self.listener_active = False
        self.current_apid = INVALID_APID

        # Listener is daemon and will exit automatically if all other threads are closed
        self.listener_thread = threading.Thread(
            target=self.__perform_operation, daemon=True
        )
        self.lock_listener = threading.Lock()
        # This Event is set by sender objects to perform mode operations
        self.event_mode_change = threading.Event()
        # This Event is set and cleared by the listener to inform the sender objects
        # if a reply has been received
        self.__event_reply_received = threading.Event()
        # This Event is set by sender objects if all necessary operations are done
        # to transition back to listener mode
        self.__event_mode_op_finished = threading.Event()

        self.__listener_mode = self.ListenerModes.LISTENER
        self.__tm_type = tm_type
        self.__queue_dict: QueueDictT = dict(
            {UNKNOWN_TARGET_ID: (deque(), self.DEFAULT_UNKNOWN_QUEUE_MAX_LEN)}
        )

    def start(self):
        if not self.event_listener_active.is_set():
            self.event_listener_active.set()
            if not self.listener_thread.is_alive():
                self.listener_thread.start()
        else:
            LOGGER.warning("TM listener is already active!")

    def stop(self):
        self.event_listener_active.clear()

    def subscribe_ccsds_tm_handler(self, apid: int, queue_max_len: int):
        if self.__tm_type == TmTypes.CCSDS_SPACE_PACKETS:
            self.__queue_dict[apid] = (deque(), queue_max_len)
        else:
            LOGGER.warning("This function only support CCSDS space packet handling")

    def set_current_apid(self, new_apid: int):
        self.current_apid = new_apid

    def set_com_if(self, com_if: CommunicationInterface):
        self.__com_if = com_if

    def is_listener_active(self) -> bool:
        return self.listener_active

    def set_timeouts(self, tm_timeout):
        self.__tm_timeout = tm_timeout

    def set_listener_mode(self, listener_mode: ListenerModes):
        if listener_mode != self.__listener_mode:
            self.event_mode_change.set()
        self.__listener_mode = listener_mode

    def reply_event(self):
        if self.__event_reply_received.is_set():
            return True
        else:
            return False

    def clear_reply_event(self):
        self.__event_reply_received.clear()

    def set_mode_op_finished(self):
        if not self.__event_mode_op_finished.is_set():
            self.__event_mode_op_finished.set()

    def ccsds_tm_received(self, apid: int = INVALID_APID):
        """This function is used to check whether any data has been received"""
        queue_dict_list = self.__queue_dict.get(apid)
        if queue_dict_list is None:
            LOGGER.warning(f"No queue available for APID {apid}")
        queue_dict = queue_dict_list[self.QUEUE_DICT_QUEUE_IDX]
        if queue_dict.__len__() > 0:
            return True
        else:
            return False

    def tm_packets_available(self):
        with acquire_timeout(
            self.lock_listener, timeout=self.DEFAULT_LOCK_TIMEOUT
        ) as acquired:
            if acquired:
                for queue_lists in self.__queue_dict.values():
                    if queue_lists[self.QUEUE_DICT_QUEUE_IDX]:
                        return True
        return False

    def retrieve_tm_packet_queues(self, clear: bool) -> QueueListT:
        queues = []
        with acquire_timeout(
            self.lock_listener, timeout=self.DEFAULT_LOCK_TIMEOUT
        ) as acquired:
            if not acquired:
                LOGGER.error("Could not acquire lock!")
            # Still continue
            for key, queue_list in self.__queue_dict.items():
                queues.append((key, queue_list[self.QUEUE_DICT_QUEUE_IDX].copy()))
            if clear:
                self.clear_tm_packet_queues(lock=False)
        return queues

    def retrieve_ccsds_tm_packet_queue(
        self, apid: int = -1, clear: bool = False
    ) -> TelemetryQueueT:
        """Retrieve the packet queue for a given APID. The TM listener will handle routing
        packets into the correct queue."""
        if apid == -1:
            apid = self.current_apid
        target_queue_list = self.__queue_dict.get(apid)
        if target_queue_list is None:
            LOGGER.warning(f"No queue available for APID {apid}")
            return deque()
        target_queue = target_queue_list[self.QUEUE_DICT_QUEUE_IDX]
        # We make sure that the queue is not manipulated while it is being copied.
        with acquire_timeout(
            self.lock_listener, timeout=self.DEFAULT_LOCK_TIMEOUT
        ) as acquired:
            if not acquired:
                LOGGER.warning(
                    f"TmListener: Blocked on lock acquisition for longer than"
                    f"{self.DEFAULT_LOCK_TIMEOUT} second!"
                )
            tm_queue_copy = target_queue.copy()
            if clear:
                target_queue.clear()
        return tm_queue_copy

    def clear_ccsds_tm_packet_queue(self, apid: int):
        if apid == -1:
            apid = self.current_apid
        target_queue = self.__queue_dict.get(apid)
        if target_queue is None:
            LOGGER.warning(f"No queue available for APID {apid}")
            return
        with acquire_timeout(
            self.lock_listener, timeout=self.DEFAULT_LOCK_TIMEOUT
        ) as acquired:
            if not acquired:
                LOGGER.warning(
                    f"TmListener: Blocked on lock acquisition for longer than"
                    f"{self.DEFAULT_LOCK_TIMEOUT} second!"
                )
            target_queue[0].clear()

    def clear_tm_packet_queues(self, lock: bool):
        locked = False
        if lock:
            locked = self.lock_listener.acquire(timeout=self.DEFAULT_LOCK_TIMEOUT)
        for queue_list in self.__queue_dict.values():
            queue_list[self.QUEUE_DICT_QUEUE_IDX].clear()
        if locked:
            self.lock_listener.release()

    def retrieve_unknown_target_queue(self):
        target_queue = self.__queue_dict.get(UNKNOWN_TARGET_ID)[
            self.QUEUE_DICT_QUEUE_IDX
        ]
        with acquire_timeout(
            self.lock_listener, timeout=self.DEFAULT_LOCK_TIMEOUT
        ) as acquired:
            if acquired:
                return target_queue.copy()

    def check_for_one_telemetry_sequence(self) -> bool:
        """Receive all telemetry for a specified time period.
        :return: True if a sequence was received
        """
        data_available = self.__com_if.data_available(timeout=0, parameters=None)
        if data_available == 0:
            return False
        elif data_available > 0:
            self.__read_telemetry_sequence()
            return True
        else:
            LOGGER.error("TmListener: Configuration error in communication interface!")
            sys.exit()

    def __perform_operation(self):
        while True:
            # This is running in a daemon thread so it will stop automatically if all other
            # threads have closed
            if self.event_listener_active.is_set():
                self.listener_active = True
                self.__default_operation()
            else:
                self.listener_active = False
                # Check every 300 ms whether connection is up again.
                time.sleep(0.3)

    def __default_operation(self):
        """Core function. Normally, polls all packets"""
        self.__perform_core_operation()
        if self.event_mode_change.is_set():
            self.event_mode_change.clear()
            start_time = time.time()
            while not self.__event_mode_op_finished.is_set():
                elapsed_time = time.time() - start_time
                if elapsed_time < TmListener.MODE_OPERATION_TIMEOUT:
                    self.__perform_mode_operation()
                else:
                    LOGGER.warning("TmListener: Mode operation timeout occured!")
                    break
            self.__event_mode_op_finished.clear()
            LOGGER.info("TmListener: Transitioning to listener mode.")
            self.__listener_mode = self.ListenerModes.LISTENER

    def __perform_core_operation(self):
        """The core operation listens for packets."""
        packet_list = self.__com_if.receive()
        if len(packet_list) > 0:
            with acquire_timeout(
                self.lock_listener, timeout=self.DEFAULT_LOCK_TIMEOUT
            ) as acquired:
                if not acquired:
                    LOGGER.warning(
                        f"TmListener: Blocked on lock acquisition for longer than"
                        f"{self.DEFAULT_LOCK_TIMEOUT} second!"
                    )
                self.__route_packets(packet_list)
            if not self.__event_reply_received.is_set():
                self.__event_reply_received.set()
        else:
            time.sleep(0.4)

    def __perform_mode_operation(self):
        """The TmListener is instructed performs certain operations based on the current
        listener mode.
        :return:
        """
        # Listener Mode
        if self.__listener_mode == self.ListenerModes.LISTENER:
            if not self.__event_mode_op_finished.is_set():
                self.__event_mode_op_finished.set()
        # Single Command Mode
        elif self.__listener_mode == self.ListenerModes.SEQUENCE:
            # This prevents the listener from listening from one more unnecessary cycle
            if self.__event_mode_op_finished.is_set():
                return
            # Listen for one reply sequence.
            if self.check_for_one_telemetry_sequence():
                # Set reply event, will be cleared by checkForFirstReply()
                if not self.__event_reply_received.is_set():
                    self.__event_reply_received.set()
            time.sleep(0.2)
        elif self.__listener_mode == self.ListenerModes.MANUAL:
            self.__perform_core_operation()

    def __read_telemetry_sequence(self):
        """Thread-safe implementation for reading a telemetry sequence."""
        start_time = time.time()
        elapsed_time = 0
        while elapsed_time < self.__tm_timeout:
            # Fast responsiveness in sequential mode
            if self.__event_mode_op_finished.is_set():
                if self.__listener_mode == self.ListenerModes.SEQUENCE:
                    return
            packets_available = self.__com_if.data_available(
                timeout=0.2, parameters=None
            )
            if packets_available > 0:
                packet_list = self.__com_if.receive()
                with acquire_timeout(
                    self.lock_listener, timeout=self.DEFAULT_LOCK_TIMEOUT
                ) as acquired:
                    if not acquired:
                        LOGGER.warning(
                            f"TmListener: Blocked on lock acquisition for longer than"
                            f"{self.DEFAULT_LOCK_TIMEOUT} second!"
                        )
                    self.__route_packets(packet_list)
            elapsed_time = time.time() - start_time
            if packets_available == 0:
                time.sleep(0.1)
        # the timeout value can be set by special TC queue entries if wiretapping_packet handling
        # takes longer, but it is reset here to the global value
        from tmtccmd.core.globals_manager import get_global
        from tmtccmd.config.definitions import CoreGlobalIds

        if self.__tm_timeout is not get_global(CoreGlobalIds.TM_TIMEOUT):
            self.__tm_timeout = get_global(CoreGlobalIds.TM_TIMEOUT)

    def __route_packets(self, tm_packet_list: TelemetryListT):
        """Route given packets. For CCSDS packets, use APID to do this"""
        for tm_packet in tm_packet_list:
            if self.__tm_type == TmTypes.CCSDS_SPACE_PACKETS:
                packet_handled = self.__handle_ccsds_space_packet(tm_packet=tm_packet)
                if packet_handled:
                    continue
            # No queue was found
            LOGGER.warning("No target queue found, inserting into unknown target queue")
            unknown_target_list = self.__queue_dict[UNKNOWN_TARGET_ID]
            unknown_target_queue = unknown_target_list[self.QUEUE_DICT_QUEUE_IDX]
            if (
                unknown_target_queue.__len__()
                > unknown_target_list[self.QUEUE_DICT_MAX_LEN_IDX]
            ):
                LOGGER.warning("Unknown target queue full. Removing oldest packet..")
                unknown_target_queue.pop()
            unknown_target_queue.appendleft(tm_packet)

    def __handle_ccsds_space_packet(self, tm_packet: bytearray) -> bool:
        if len(tm_packet) < 6:
            LOGGER.warning("TM packet to small to be a CCSDS space packet")
        else:
            apid = get_apid_from_raw_space_packet(raw_packet=tm_packet)
            target_queue_list = self.__queue_dict.get(apid)
            if target_queue_list is None:
                LOGGER.warning(f"No TM handler assigned for APID {apid}")
            else:
                target_queue = target_queue_list[self.QUEUE_DICT_QUEUE_IDX]
                if (
                    target_queue.__len__()
                    > target_queue_list[self.QUEUE_DICT_MAX_LEN_IDX]
                ):
                    LOGGER.warning(
                        f"Target queue for APID {apid} full. Removing oldest packet.."
                    )
                    target_queue.pop()
                target_queue.appendleft(tm_packet)
                return True
        return False



import sys
from typing import Optional, Tuple

from tmtccmd.config.definitions import CoreGlobalIds, CoreComInterfaces
from tmtccmd.core.globals_manager import get_global, update_global
from tmtccmd.com_if.com_interface_base import CommunicationInterface
from tmtccmd.com_if.serial_com_if import (
    SerialConfigIds,
    SerialCommunicationType,
    SerialComIF,
)
from tmtccmd.com_if.serial_utilities import determine_com_port, determine_baud_rate
from tmtccmd.com_if.tcpip_utilities import TcpIpConfigIds, TcpIpType
from tmtccmd.utility.logger import get_console_logger
from tmtccmd.utility.tmtc_printer import TmTcPrinter
from tmtccmd.com_if.tcpip_udp_com_if import TcpIpUdpComIF
from tmtccmd.com_if.tcpip_tcp_com_if import TcpIpTcpComIF, TcpCommunicationType

LOGGER = get_console_logger()


def create_communication_interface_default(
    com_if_key: str,
    tmtc_printer: TmTcPrinter,
    json_cfg_path: str,
    space_packet_ids: Tuple[int] = (0,),
) -> Optional[CommunicationInterface]:
    """Return the desired communication interface object

    :param com_if_key:
    :param tmtc_printer:
    :param json_cfg_path:
    TODO: List of packet IDs?
    :param space_packet_id: Can be used by communication interfaces as a start marker (e.g. TCP)
    :return:
    """
    from tmtccmd.com_if.dummy_com_if import DummyComIF
    from tmtccmd.com_if.qemu_com_if import QEMUComIF

    try:
        if (
            com_if_key == CoreComInterfaces.TCPIP_UDP.value
            or com_if_key == CoreComInterfaces.TCPIP_TCP.value
        ):
            communication_interface = create_default_tcpip_interface(
                com_if_key=com_if_key,
                json_cfg_path=json_cfg_path,
                tmtc_printer=tmtc_printer,
                space_packet_ids=space_packet_ids,
            )
        elif (
            com_if_key == CoreComInterfaces.SERIAL_DLE.value
            or com_if_key == CoreComInterfaces.SERIAL_FIXED_FRAME.value
        ):
            communication_interface = create_default_serial_interface(
                com_if_key=com_if_key,
                tmtc_printer=tmtc_printer,
                json_cfg_path=json_cfg_path,
            )
        elif com_if_key == CoreComInterfaces.SERIAL_QEMU.value:
            serial_cfg = get_global(CoreGlobalIds.SERIAL_CONFIG)
            serial_timeout = serial_cfg[SerialConfigIds.SERIAL_TIMEOUT]
            communication_interface = QEMUComIF(
                tmtc_printer=tmtc_printer,
                serial_timeout=serial_timeout,
                ser_com_type=SerialCommunicationType.DLE_ENCODING,
            )
            dle_max_queue_len = serial_cfg[SerialConfigIds.SERIAL_DLE_QUEUE_LEN]
            dle_max_frame_size = serial_cfg[SerialConfigIds.SERIAL_DLE_MAX_FRAME_SIZE]
            communication_interface.set_dle_settings(
                dle_max_queue_len, dle_max_frame_size, serial_timeout
            )
        else:
            communication_interface = DummyComIF(
                com_if_key=com_if_key, tmtc_printer=tmtc_printer
            )
        if communication_interface is None:
            return communication_interface
        if not communication_interface.valid:
            LOGGER.warning("Invalid communication interface!")
            return None
        communication_interface.initialize()
        return communication_interface
    except ConnectionRefusedError:
        LOGGER.exception(f"TCP/IP connection refused")
        if com_if_key == CoreComInterfaces.TCPIP_UDP.value:
            LOGGER.warning("Make sure that a UDP server is running")
        if com_if_key == CoreComInterfaces.TCPIP_TCP.value:
            LOGGER.warning("Make sure that a TCP server is running")
        sys.exit(1)
    except (IOError, OSError):
        LOGGER.exception(f"Error setting up communication interface")
        sys.exit(1)


def default_tcpip_cfg_setup(
    tcpip_type: TcpIpType, json_cfg_path: str, space_packet_ids: Tuple[int] = (0,)
):
    """Default setup for TCP/IP communication interfaces. This intantiates all required data in the
    globals manager so a TCP/IP communication interface can be built with
    :func:`create_default_tcpip_interface`
    :param tcpip_type:
    :param json_cfg_path:
    :param space_packet_id:       Required if the TCP com interface needs to parse space packets
    :return:
    """
    from tmtccmd.com_if.tcpip_utilities import (
        determine_udp_send_address,
        determine_tcp_send_address,
        determine_recv_buffer_len,
    )

    update_global(CoreGlobalIds.USE_ETHERNET, True)
    recv_tuple = None
    if tcpip_type == TcpIpType.UDP:
        send_tuple = determine_udp_send_address(json_cfg_path=json_cfg_path)
    elif tcpip_type == TcpIpType.TCP:
        send_tuple = determine_tcp_send_address(json_cfg_path=json_cfg_path)
    else:
        send_tuple = ()
    max_recv_buf_size = determine_recv_buffer_len(
        json_cfg_path=json_cfg_path, tcpip_type=tcpip_type
    )
    ethernet_cfg_dict = get_global(CoreGlobalIds.ETHERNET_CONFIG)
    ethernet_cfg_dict.update({TcpIpConfigIds.SEND_ADDRESS: send_tuple})
    ethernet_cfg_dict.update({TcpIpConfigIds.RECV_ADDRESS: recv_tuple})
    ethernet_cfg_dict.update({TcpIpConfigIds.RECV_MAX_SIZE: max_recv_buf_size})
    if space_packet_ids == (0,) and tcpip_type == TcpIpType.TCP:
        LOGGER.warning(
            "TCP communication interface without any specified space packet ID might not work!"
        )
    ethernet_cfg_dict.update({TcpIpConfigIds.SPACE_PACKET_ID: space_packet_ids})
    update_global(CoreGlobalIds.ETHERNET_CONFIG, ethernet_cfg_dict)


def default_serial_cfg_setup(com_if_key: str, json_cfg_path: str):
    """Default setup for serial interfaces

    :param com_if_key:
    :param json_cfg_path:
    :return:
    """
    baud_rate = determine_baud_rate(json_cfg_path=json_cfg_path)
    serial_port = determine_com_port(json_cfg_path=json_cfg_path)
    set_up_serial_cfg(
        json_cfg_path=json_cfg_path,
        com_if_key=com_if_key,
        baud_rate=baud_rate,
        com_port=serial_port,
    )


def create_default_tcpip_interface(
    com_if_key: str,
    tmtc_printer: TmTcPrinter,
    json_cfg_path: str,
    space_packet_ids: Tuple[int] = (0,),
) -> Optional[CommunicationInterface]:
    """Create a default serial interface. Requires a certain set of global variables set up. See
    :func:`default_tcpip_cfg_setup` for more details.

    :param com_if_key:
    :param tmtc_printer:
    :param json_cfg_path:
    :param space_packet_ids: Two byte packet IDs which is used by TCP to parse space packets
    :return:
    """
    communication_interface = None
    if com_if_key == CoreComInterfaces.TCPIP_UDP.value:
        default_tcpip_cfg_setup(tcpip_type=TcpIpType.UDP, json_cfg_path=json_cfg_path)
    elif com_if_key == CoreComInterfaces.TCPIP_TCP.value:
        default_tcpip_cfg_setup(
            tcpip_type=TcpIpType.TCP,
            json_cfg_path=json_cfg_path,
            space_packet_ids=space_packet_ids,
        )
    ethernet_cfg_dict = get_global(CoreGlobalIds.ETHERNET_CONFIG)
    send_addr = ethernet_cfg_dict[TcpIpConfigIds.SEND_ADDRESS]
    recv_addr = ethernet_cfg_dict[TcpIpConfigIds.RECV_ADDRESS]
    max_recv_size = ethernet_cfg_dict[TcpIpConfigIds.RECV_MAX_SIZE]
    init_mode = get_global(CoreGlobalIds.MODE)
    space_packet_id = ethernet_cfg_dict[TcpIpConfigIds.SPACE_PACKET_ID]
    if com_if_key == CoreComInterfaces.TCPIP_UDP.value:
        communication_interface = TcpIpUdpComIF(
            com_if_key=com_if_key,
            tm_timeout=get_global(CoreGlobalIds.TM_TIMEOUT),
            tc_timeout_factor=get_global(CoreGlobalIds.TC_SEND_TIMEOUT_FACTOR),
            send_address=send_addr,
            recv_addr=recv_addr,
            max_recv_size=max_recv_size,
            tmtc_printer=tmtc_printer,
            init_mode=init_mode,
        )
    elif com_if_key == CoreComInterfaces.TCPIP_TCP.value:
        communication_interface = TcpIpTcpComIF(
            com_if_key=com_if_key,
            com_type=TcpCommunicationType.SPACE_PACKETS,
            space_packet_ids=space_packet_ids,
            tm_timeout=get_global(CoreGlobalIds.TM_TIMEOUT),
            tc_timeout_factor=get_global(CoreGlobalIds.TC_SEND_TIMEOUT_FACTOR),
            tm_polling_freqency=0.5,
            send_address=send_addr,
            max_recv_size=max_recv_size,
            tmtc_printer=tmtc_printer,
            init_mode=init_mode,
        )
    return communication_interface


def create_default_serial_interface(
    com_if_key: str, tmtc_printer: TmTcPrinter, json_cfg_path: str
) -> Optional[CommunicationInterface]:
    """Create a default serial interface. Requires a certain set of global variables set up. See
    :func:`set_up_serial_cfg` for more details.

    :param com_if_key:
    :param tmtc_printer:
    :param json_cfg_path:
    :return:
    """
    try:
        # For a serial communication interface, there are some configuration values like
        # baud rate and serial port which need to be set once but are expected to stay
        # the same for a given machine. Therefore, we use a JSON file to store and extract
        # those values
        if (
            com_if_key == CoreComInterfaces.SERIAL_DLE.value
            or com_if_key == CoreComInterfaces.SERIAL_FIXED_FRAME.value
            or com_if_key == CoreComInterfaces.SERIAL_QEMU.value
        ):
            default_serial_cfg_setup(com_if_key=com_if_key, json_cfg_path=json_cfg_path)
        serial_cfg = get_global(CoreGlobalIds.SERIAL_CONFIG)
        serial_baudrate = serial_cfg[SerialConfigIds.SERIAL_BAUD_RATE]
        serial_timeout = serial_cfg[SerialConfigIds.SERIAL_TIMEOUT]
        com_port = serial_cfg[SerialConfigIds.SERIAL_PORT]
        if com_if_key == CoreComInterfaces.SERIAL_DLE.value:
            ser_com_type = SerialCommunicationType.DLE_ENCODING
        else:
            ser_com_type = SerialCommunicationType.FIXED_FRAME_BASED
        communication_interface = SerialComIF(
            com_if_key=com_if_key,
            tmtc_printer=tmtc_printer,
            com_port=com_port,
            baud_rate=serial_baudrate,
            serial_timeout=serial_timeout,
            ser_com_type=ser_com_type,
        )
        if com_if_key == CoreComInterfaces.SERIAL_DLE:
            dle_max_queue_len = serial_cfg[SerialConfigIds.SERIAL_DLE_QUEUE_LEN]
            dle_max_frame_size = serial_cfg[SerialConfigIds.SERIAL_DLE_MAX_FRAME_SIZE]
            communication_interface.set_dle_settings(
                dle_max_queue_len, dle_max_frame_size, serial_timeout
            )
    except KeyError:
        LOGGER.warning("Serial configuration global not configured properly")
        return None
    return communication_interface


def set_up_serial_cfg(
    json_cfg_path: str,
    com_if_key: str,
    baud_rate: int,
    com_port: str = "",
    tm_timeout: float = 0.01,
    ser_com_type: SerialCommunicationType = SerialCommunicationType.DLE_ENCODING,
    ser_frame_size: int = 256,
    dle_queue_len: int = 25,
    dle_frame_size: int = 1024,
):
    """Default configuration to set up serial communication. The serial port and the baud rate
    will be determined from a JSON configuration file and prompted from the user. Sets up all
    global variables so that a serial communication interface can be built with
    :func:`create_default_serial_interface`
    :param json_cfg_path:
    :param com_if_key:
    :param com_port:
    :param baud_rate:
    :param tm_timeout:
    :param ser_com_type:
    :param ser_frame_size:
    :param dle_queue_len:
    :param dle_frame_size:
    :return:
    """
    update_global(CoreGlobalIds.USE_SERIAL, True)
    if (
        com_if_key == CoreComInterfaces.SERIAL_DLE.value
        or com_if_key == CoreComInterfaces.SERIAL_FIXED_FRAME.value
    ) and com_port == "":
        LOGGER.warning("Invalid serial port specified!")
        com_port = determine_com_port(json_cfg_path=json_cfg_path)
    serial_cfg_dict = get_global(CoreGlobalIds.SERIAL_CONFIG)
    serial_cfg_dict.update({SerialConfigIds.SERIAL_PORT: com_port})
    serial_cfg_dict.update({SerialConfigIds.SERIAL_BAUD_RATE: baud_rate})
    serial_cfg_dict.update({SerialConfigIds.SERIAL_TIMEOUT: tm_timeout})
    serial_cfg_dict.update({SerialConfigIds.SERIAL_COMM_TYPE: ser_com_type})
    serial_cfg_dict.update({SerialConfigIds.SERIAL_FRAME_SIZE: ser_frame_size})
    serial_cfg_dict.update({SerialConfigIds.SERIAL_DLE_QUEUE_LEN: dle_queue_len})
    serial_cfg_dict.update({SerialConfigIds.SERIAL_DLE_MAX_FRAME_SIZE: dle_frame_size})
    update_global(CoreGlobalIds.SERIAL_CONFIG, serial_cfg_dict)



import enum


class EcssPtc(enum.IntEnum):
    BOOLEAN = 1
    ENUMERATED = 2
    UNSIGNED = 3
    SIGNED = 4
    # Float or double values
    REAL = 5
    BIT_STRING = 6
    OCTET_STRING = 7
    CHARACTER_STRING = 8
    ABSOLUTE_TIME = 9
    RELATIVE_TIME = 10
    DEDUCED = 11
    PACKET = 12


class EcssPfcUnsigned(enum.IntEnum):
    FOUR_BIT = 0
    FIVE_BIT = 1
    SIX_BIT = 2
    SEVEN_BIT = 3
    ONE_BYTE = 4
    NINE_BIT = 5
    TEN_BIT = 6
    ELEVEN_BIT = 7
    TWELVE_BIT = 8
    THIRTEEN_BIT = 9
    FOURTEEN_BIT = 10
    FIFTEEN_BIT = 11
    TWO_BYTES = 12
    THREE_BYTES = 13
    FOUR_BYTES = 14
    SIX_BYTES = 15
    EIGHT_BYTES = 16
    ONE_BIT = 17
    TWO_BIT = 18
    THREE_BIT = 19


class EcssPfcSigned(enum.IntEnum):
    FOUR_BIT = 0
    FIVE_BIT = 1
    SIX_BIT = 2
    SEVEN_BIT = 3
    ONE_BYTE = 4
    NINE_BIT = 5
    TEN_BIT = 6
    ELEVEN_BIT = 7
    TWELVE_BIT = 8
    THIRTEEN_BIT = 9
    FOURTEEN_BIT = 10
    FIFTEEN_BIT = 11
    TWO_BYTES = 12
    THREE_BYTES = 13
    FOUR_BYTES = 14
    SIX_BYTES = 15
    EIGHT_BYTES = 16


class EcssPfcReal(enum.IntEnum):
    FLOAT_SIMPLE_PRECISION_IEEE = 1
    DOUBLE_PRECISION_IEEE = 2
    FLOAT_PRECISION_MIL_STD_4_OCTETS = 3
    DOUBLE_PRECISION_MIL_STD_6_OCTETS = 4


class CustomSubservices(enum.IntEnum):
    LOAD = (128,)
    DUMP = 129



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../src"))
import tmtccmd

# -- Project information -----------------------------------------------------

project = "tmtccmd"
copyright = "2021, Robin Mueller"
author = "Robin Mueller"

# The full version, including alpha/beta/rc tags
version = release = tmtccmd.__version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx.ext.autodoc", "sphinx.ext.intersphinx"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_trees = ["_build"]

intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "logo_tmtccmd.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# -- Options for LaTeX output --------------------------------------------------

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = "logo_tmtccmd.png"



from tmtccmd.tm.definitions import TmTypes
from spacepackets.ecss.tm import PusTelemetry
from tmtccmd.tm.service_5_event import Service5TM
from tmtccmd.pus.service_1_verification import Service1TM
from tmtccmd.pus.service_17_test import Service17TMExtended
from tmtccmd.utility.logger import get_console_logger
from tmtccmd.utility.tmtc_printer import TmTcPrinter

LOGGER = get_console_logger()


class TmHandler:
    def __init__(self, tm_type: TmTypes):
        self._tm_type = tm_type

    def get_type(self):
        return self._tm_type


def default_ccsds_packet_handler(
    apid: int, raw_tm_packet: bytearray, tmtc_printer: TmTcPrinter
):
    """Default implementation only prints the packet"""
    default_factory_hook(raw_tm_packet=raw_tm_packet, tmtc_printer=tmtc_printer)


def default_factory_hook(raw_tm_packet: bytearray, tmtc_printer: TmTcPrinter):
    service_type = raw_tm_packet[7]
    tm_packet = None
    if service_type == 1:
        tm_packet = Service17TMExtended.unpack(raw_telemetry=raw_tm_packet)
    if service_type == 5:
        tm_packet = Service5TM.unpack(raw_telemetry=raw_tm_packet)
    if service_type == 17:
        tm_packet = Service17TMExtended.unpack(raw_telemetry=raw_tm_packet)
    if tm_packet is None:
        LOGGER.info(
            f"The service {service_type} is not implemented in Telemetry Factory"
        )
        tm_packet = PusTelemetry.unpack(raw_telemetry=raw_tm_packet)
    tmtc_printer.print_telemetry(packet_if=tm_packet, info_if=tm_packet)



import argparse
import collections.abc
import pprint
from typing import Union, List, Dict

from tmtccmd.utility.logger import get_console_logger
from tmtccmd.utility.conf_util import check_args_in_dict, print_core_globals
from spacepackets.ecss.conf import (
    PusVersion,
    set_default_tc_apid,
    set_default_tm_apid,
    set_pus_tc_version,
    set_pus_tm_version,
)
from tmtccmd.core.globals_manager import update_global, get_global
from tmtccmd.config.definitions import (
    CoreGlobalIds,
    CoreModeList,
    CoreServiceList,
    CoreModeStrings,
    CoreComInterfacesDict,
    CoreComInterfaces,
)
from tmtccmd.com_if.com_if_utilities import determine_com_if
from tmtccmd.config.definitions import (
    DEBUG_MODE,
    ServiceOpCodeDictT,
    OpCodeDictKeys,
    ComIFDictT,
    ServiceDictValueT,
    OpCodeEntryT,
    OpCodeInfoT,
    OpCodeOptionsT,
    OpCodeNameT,
)

LOGGER = get_console_logger()
SERVICE_OP_CODE_DICT = dict()


def set_json_cfg_path(json_cfg_path: str):
    update_global(CoreGlobalIds.JSON_CFG_PATH, json_cfg_path)


def get_json_cfg_path() -> str:
    return get_global(CoreGlobalIds.JSON_CFG_PATH)


def set_glob_com_if_dict(custom_com_if_dict: ComIFDictT):
    CoreComInterfacesDict.update(custom_com_if_dict)
    update_global(CoreGlobalIds.COM_IF_DICT, CoreComInterfacesDict)


def get_glob_com_if_dict() -> ComIFDictT:
    return get_global(CoreGlobalIds.COM_IF_DICT)


def set_default_globals_pre_args_parsing(
    gui: bool,
    tc_apid: int,
    tm_apid: int,
    pus_tc_version: PusVersion = PusVersion.PUS_C,
    pus_tm_version: PusVersion = PusVersion.PUS_C,
    com_if_id: str = CoreComInterfaces.DUMMY.value,
    custom_com_if_dict=None,
    display_mode="long",
    tm_timeout: float = 4.0,
    print_to_file: bool = True,
    tc_send_timeout_factor: float = 2.0,
):
    if custom_com_if_dict is None:
        custom_com_if_dict = dict()
    set_default_tc_apid(tc_apid=tc_apid)
    set_default_tm_apid(tm_apid=tm_apid)
    set_pus_tc_version(pus_tc_version)
    set_pus_tm_version(pus_tm_version)
    update_global(CoreGlobalIds.COM_IF, com_if_id)
    update_global(CoreGlobalIds.TC_SEND_TIMEOUT_FACTOR, tc_send_timeout_factor)
    update_global(CoreGlobalIds.TM_TIMEOUT, tm_timeout)
    update_global(CoreGlobalIds.DISPLAY_MODE, display_mode)
    update_global(CoreGlobalIds.PRINT_TO_FILE, print_to_file)
    update_global(CoreGlobalIds.CURRENT_SERVICE, CoreServiceList.SERVICE_17.value)
    update_global(CoreGlobalIds.SERIAL_CONFIG, dict())
    update_global(CoreGlobalIds.ETHERNET_CONFIG, dict())
    set_glob_com_if_dict(custom_com_if_dict=custom_com_if_dict)
    pp = pprint.PrettyPrinter()
    update_global(CoreGlobalIds.PRETTY_PRINTER, pp)
    update_global(CoreGlobalIds.TM_LISTENER_HANDLE, None)
    update_global(CoreGlobalIds.COM_INTERFACE_HANDLE, None)
    update_global(CoreGlobalIds.TMTC_PRINTER_HANDLE, None)
    update_global(CoreGlobalIds.PRINT_RAW_TM, False)
    update_global(CoreGlobalIds.USE_LISTENER_AFTER_OP, True)
    update_global(CoreGlobalIds.RESEND_TC, False)
    update_global(CoreGlobalIds.OP_CODE, "0")
    update_global(CoreGlobalIds.MODE, CoreModeList.LISTENER_MODE)


def set_default_globals_post_args_parsing(
    args: argparse.Namespace,
    json_cfg_path: str,
    custom_modes_list: Union[None, List[Union[collections.abc.Iterable, dict]]] = None,
    custom_services_list: Union[
        None, List[Union[collections.abc.Iterable, dict]]
    ] = None,
    custom_com_if_dict: Dict[str, any] = None,
):
    """This function takes the argument namespace as a parameter and determines
    a set of globals from the parsed arguments.
    If custom dictionaries are specified, the developer should take care of specifying
    integers as keys and the string representation of the command line argument as value.
    This will be used for internalization.

    :param args: Namespace generated by parsing command line arguments.
    :param json_cfg_path:
    :param custom_modes_list: List of collections or dictionaries containing custom modes
    :param custom_services_list: List of collections or dictionaries containing custom services
    :param custom_com_if_dict: List of collections or dictionaries containing customcommunication interfaces
    :return:
    """

    handle_mode_arg(args=args, custom_modes_list=custom_modes_list)
    handle_com_if_arg(
        args=args, json_cfg_path=json_cfg_path, custom_com_if_dict=custom_com_if_dict
    )

    display_mode_param = "long"
    if args.short_display_mode is not None:
        if args.short_display_mode:
            display_mode_param = "short"
        else:
            display_mode_param = "long"
    update_global(CoreGlobalIds.DISPLAY_MODE, display_mode_param)

    try:
        service_param = args.service
    except AttributeError:
        LOGGER.warning(
            "Passed namespace does not contain the service (-s) argument. "
            "Setting test service ID (17)"
        )
        service_param = CoreServiceList.SERVICE_17.value
    update_global(CoreGlobalIds.CURRENT_SERVICE, service_param)
    # Not used for now
    """
    check_and_set_core_service_arg(
        service_arg=service_param, custom_service_list=custom_services_list
    )
    """

    if args.op_code is None:
        op_code = 0
    else:
        op_code = str(args.op_code).lower()
    update_global(CoreGlobalIds.OP_CODE, op_code)

    try:
        check_and_set_other_args(args=args)
    except AttributeError:
        LOGGER.exception("Passed arguments are missing components.")

    if DEBUG_MODE:
        print_core_globals()


def handle_mode_arg(
    args,
    custom_modes_list: Union[None, List[Union[collections.abc.Iterable, dict]]] = None,
) -> int:
    # Determine communication interface from arguments. Must be contained in core modes list
    try:
        mode_param = args.mode
    except AttributeError:
        LOGGER.warning("Passed namespace does not contain the mode (-m) argument")
        mode_param = CoreModeList.LISTENER_MODE
    mode_param = check_and_set_core_mode_arg(
        mode_arg=mode_param, custom_modes_list=custom_modes_list
    )
    update_global(CoreGlobalIds.MODE, mode_param)
    return mode_param


def handle_com_if_arg(
    args, json_cfg_path: str, custom_com_if_dict: Dict[str, any] = None
):
    all_com_ifs = CoreComInterfacesDict
    if custom_com_if_dict is not None:
        all_com_ifs = CoreComInterfacesDict.update(custom_com_if_dict)
    try:
        com_if_key = str(args.com_if)
    except AttributeError:
        LOGGER.warning("No communication interface specified")
        LOGGER.warning("Trying to set from existing configuration..")
        com_if_key = determine_com_if(
            com_if_dict=all_com_ifs, json_cfg_path=json_cfg_path
        )
    if com_if_key == CoreComInterfaces.UNSPECIFIED.value:
        com_if_key = determine_com_if(
            com_if_dict=all_com_ifs, json_cfg_path=json_cfg_path
        )
    update_global(CoreGlobalIds.COM_IF, com_if_key)
    try:
        LOGGER.info(f"Communication interface: {all_com_ifs[com_if_key]}")
    except KeyError as e:
        LOGGER.error(f"Invalid communication interface key {com_if_key}, error {e}")


def check_and_set_other_args(args):
    if args.listener is not None:
        update_global(CoreGlobalIds.USE_LISTENER_AFTER_OP, args.listener)
    if args.tm_timeout is not None:
        update_global(CoreGlobalIds.TM_TIMEOUT, args.tm_timeout)
    if args.print_hk is not None:
        update_global(CoreGlobalIds.PRINT_HK, args.print_hk)
    if args.print_tm is not None:
        update_global(CoreGlobalIds.PRINT_TM, args.print_tm)
    if args.raw_data_print is not None:
        update_global(CoreGlobalIds.PRINT_RAW_TM, args.raw_data_print)
    if args.print_log is not None:
        update_global(CoreGlobalIds.PRINT_TO_FILE, args.print_log)
    if args.resend_tc is not None:
        update_global(CoreGlobalIds.RESEND_TC, args.resend_tc)
    update_global(CoreGlobalIds.TC_SEND_TIMEOUT_FACTOR, 3)


def check_and_set_core_mode_arg(
    mode_arg: any,
    custom_modes_list: Union[None, List[Union[dict, collections.abc.Iterable]]] = None,
) -> int:
    """Checks whether the mode argument is contained inside the core mode list integer enumeration
    or a custom mode list integer which can be passed optionally.
    This function will set the single command mode as the global mode parameter if the passed mode
    is not found in either enumerations.

    :param mode_arg:
    :param custom_modes_list:
    :return: Mode value which was set
    """
    in_enum, mode_value = check_args_in_dict(
        param=mode_arg, iterable=CoreModeList, warning_hint="mode integers"
    )
    if not in_enum:
        in_enum, mode_value = check_args_in_dict(
            param=mode_arg, iterable=CoreModeStrings, warning_hint="mode strings"
        )
    if in_enum:
        update_global(CoreGlobalIds.MODE, mode_value)
        return mode_value

    mode_arg_invalid = False
    if custom_modes_list is not None:
        for custom_mode_entry in custom_modes_list:
            in_enum, mode_value = check_args_in_dict(
                param=mode_arg, iterable=custom_mode_entry, warning_hint="custom mode"
            )
            if in_enum:
                break
        if not in_enum:
            mode_arg_invalid = True
    else:
        mode_arg_invalid = True

    if mode_arg_invalid:
        LOGGER.warning(
            f"Passed mode argument might be invalid, "
            f"setting to {CoreModeList.SEQUENTIAL_CMD_MODE}"
        )
        mode_value = CoreModeList.SEQUENTIAL_CMD_MODE
    update_global(CoreGlobalIds.MODE, mode_value)
    return mode_value


def check_and_set_core_service_arg(
    service_arg: any, custom_service_list: collections.abc.Iterable = None
):
    in_enum, service_value = check_args_in_dict(
        param=service_arg, iterable=CoreServiceList, warning_hint="service"
    )
    if in_enum:
        update_global(CoreGlobalIds.CURRENT_SERVICE, service_value)
        return

    service_arg_invalid = False
    if custom_service_list is not None:
        for custom_services_entry in custom_service_list:
            in_enum, service_value = check_args_in_dict(
                param=service_arg,
                iterable=custom_services_entry,
                warning_hint="custom mode",
            )
            if in_enum:
                break
        if not in_enum:
            service_arg_invalid = True
    else:
        service_arg_invalid = True

    if service_arg_invalid:
        LOGGER.warning(
            f"Passed service argument might be invalid, "
            f"setting to {CoreServiceList.SERVICE_17}"
        )
        service_value = CoreServiceList.SERVICE_17
    update_global(CoreGlobalIds.CURRENT_SERVICE, service_value)


def get_default_service_op_code_dict() -> ServiceOpCodeDictT:
    global SERVICE_OP_CODE_DICT
    service_op_code_dict = SERVICE_OP_CODE_DICT
    if service_op_code_dict == dict():
        op_code_dict_srv_5 = {
            "0": ("Event Test", {OpCodeDictKeys.TIMEOUT: 2.0}),
        }
        service_5_tuple = ("PUS Service 5 Event", op_code_dict_srv_5)
        op_code_dict_srv_17 = {
            "0": ("Ping Test", {OpCodeDictKeys.TIMEOUT: 2.2}),
        }
        service_17_tuple = ("PUS Service 17 Test", op_code_dict_srv_17)

        service_op_code_dict[CoreServiceList.SERVICE_5.value] = service_5_tuple
        service_op_code_dict[CoreServiceList.SERVICE_17.value] = service_17_tuple
        # SERVICE_OP_CODE_DICT = service_op_code_dict
    return service_op_code_dict


def add_op_code_entry(
    op_code_dict: OpCodeEntryT,
    keys: OpCodeNameT,
    info: str,
    options: OpCodeOptionsT = None,
):
    if isinstance(keys, str):
        keys = [keys]
    op_code_dict.update(OpCodeEntryT.fromkeys(keys, (info, options)))


def add_service_op_code_entry(
    srv_op_code_dict: ServiceOpCodeDictT,
    name: str,
    info: str,
    op_code_entry: OpCodeEntryT,
):
    srv_op_code_dict.update({name: (info, op_code_entry)})



from __future__ import annotations
import struct

from spacepackets.ecss.definitions import PusServices
from spacepackets.ecss.tm import CdsShortTimestamp, PusVersion, PusTelemetry

from tmtccmd.tm.base import PusTmInfoBase, PusTmBase
from tmtccmd.utility.logger import get_console_logger

LOGGER = get_console_logger()


class FileInfo:
    def __init__(
        self,
        file_name: str,
        repo_path: str,
        file_size: int = 0,
        lock_status: bool = False,
    ):
        self.file_name = file_name
        self.repo_path = repo_path
        self.file_size = file_size
        self.lock_status = lock_status


class Service23TM(PusTmInfoBase, PusTmBase):
    MAX_REPOSITORY_LENGTH = 64
    MAX_FILENAME_LENGTH = 12

    def __init__(
        self,
        subservice_id: int,
        object_id: bytearray,
        repo_path: str,
        file_name: str,
        time: CdsShortTimestamp = None,
        ssc: int = 0,
        source_data: bytearray = bytearray([]),
        apid: int = -1,
        packet_version: int = 0b000,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
        pus_tm_version: int = 0b0001,
        ack: int = 0b1111,
        secondary_header_flag: bool = True,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        self.object_id = object_id
        self.data_start_idx = 0
        self.file_info = FileInfo(file_name=file_name, repo_path=repo_path)
        pus_tm = PusTelemetry(
            service=PusServices.SERVICE_23_FILE_MGMT,
            subservice=subservice_id,
            time=time,
            ssc=ssc,
            source_data=source_data,
            apid=apid,
            packet_version=packet_version,
            pus_version=pus_version,
            pus_tm_version=pus_tm_version,
            ack=ack,
            secondary_header_flag=secondary_header_flag,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )
        PusTmBase.__init__(self, pus_tm=pus_tm)
        PusTmInfoBase.__init__(self, pus_tm=pus_tm)

        self.specify_packet_info("File Service Reply")

    @staticmethod
    def __init_without_base(instance: Service23TM):
        tm_data = instance.get_tm_data()
        if len(tm_data) < 4:
            LOGGER.error("Service23TM: Invalid packet format!")
            return
        instance.object_id = struct.unpack("!I", tm_data[0:4])[0]
        instance.file_info.file_size = 0
        instance.file_info.lock_status = False
        instance.data_start_idx = 0
        if instance.get_subservice() == 4:
            instance.unpack_repo_and_filename()
            instance.unpack_file_attributes()
        elif instance.get_subservice() == 132:
            instance.unpack_repo_and_filename()

    @classmethod
    def __empty(cls) -> Service23TM:
        return cls(subservice_id=-1, object_id=bytearray(4), repo_path="", file_name="")

    @classmethod
    def unpack(
        cls,
        raw_telemetry: bytearray,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
    ) -> Service23TM:
        service_23_tm = cls.__empty()
        service_23_tm.pus_tm = PusTelemetry.unpack(
            raw_telemetry=raw_telemetry, pus_version=pus_version
        )
        service_23_tm.__init_without_base(instance=service_23_tm)
        return service_23_tm

    def unpack_repo_and_filename(self):
        tm_data = self.get_tm_data()
        repo_path_found = False
        path_idx_start = 0
        max_len_to_scan = len(self.get_tm_data()) - 4
        for idx in range(4, max_len_to_scan):
            if not repo_path_found and tm_data[idx] == 0:
                repo_bytes = tm_data[4:idx]
                self.file_info.repo_path = repo_bytes.decode("utf-8")
                path_idx_start = idx + 1
                idx += 1
                repo_path_found = True
            if repo_path_found:
                if tm_data[idx] == 0:
                    filename_bytes = tm_data[path_idx_start:idx]
                    self.file_info.file_name = filename_bytes.decode("utf-8")
                    self.data_start_idx = idx + 1
                    break

    def unpack_file_attributes(self):
        # Size of file length (4) + lock status (1), adapt if more field are added!
        print(len(self.get_tm_data()) - self.data_start_idx)
        if len(self.get_tm_data()) - self.data_start_idx != 5:
            LOGGER.error("Service23TM: Invalid lenght of file attributes data")
            return
        self.file_info.file_size = struct.unpack(
            "!I", self.get_tm_data()[self.data_start_idx : self.data_start_idx + 4]
        )[0]
        self.file_info.lock_status = self.get_tm_data()[self.data_start_idx + 4]

    def append_telemetry_content(self, content_list: list):
        super().append_telemetry_content(content_list)
        content_list.append(self.object_id)
        content_list.append(self.file_info.repo_path)
        content_list.append(self.file_info.file_name)
        if self.get_subservice() == 4:
            content_list.append(self.file_info.file_size)
            if self.file_info.lock_status == 0:
                content_list.append("No")
            else:
                content_list.append("Yes")

    def append_telemetry_column_headers(self, header_list: list):
        super().append_telemetry_column_headers(header_list)
        header_list.append("Object ID")
        header_list.append("Repo Path")
        header_list.append("File Name")
        if self.get_subservice() == 4:
            header_list.append("File Size")
            header_list.append("Locked")



# -*- coding: utf-8 -*-
"""PUS Service 3 components
"""
from __future__ import annotations
from abc import abstractmethod
import struct

from spacepackets.ecss.tm import CdsShortTimestamp, PusVersion, PusTelemetry
from tmtccmd.tm.base import PusTmInfoBase, PusTmBase
from tmtccmd.tm.service_3_base import Service3Base
from tmtccmd.utility.logger import get_console_logger
from typing import Tuple, List


LOGGER = get_console_logger()


class Service3TM(Service3Base, PusTmBase, PusTmInfoBase):
    """This class encapsulates the format of Service 3 telemetry
    This class was written to handle Service 3 telemetry coming from the on-board software
    based on the Flight Software Framework (FSFW). A custom class can be defined, but should then
    implement Service3Base.
    """

    # Minimal packet contains SID, which consists of object ID(4) and set ID(4)
    DEFAULT_MINIMAL_PACKET_SIZE = 8
    # Minimal structure report contains SID (8), reporting status(1), validity flag (1),
    # collection interval as float (4) and number of parameters(1)
    STRUCTURE_REPORT_FIXED_HEADER_SIZE = DEFAULT_MINIMAL_PACKET_SIZE + 7

    def __init__(
        self,
        subservice_id: int,
        time: CdsShortTimestamp,
        hk_data: bytearray,
        custom_hk_handling: bool = False,
        ssc: int = 0,
        apid: int = -1,
        minimum_reply_size: int = DEFAULT_MINIMAL_PACKET_SIZE,
        minimum_structure_report_header_size: int = STRUCTURE_REPORT_FIXED_HEADER_SIZE,
        packet_version: int = 0b000,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
        pus_tm_version: int = 0b0001,
        secondary_header_flag: bool = True,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        """Service 3 packet class representation which can be built from a raw bytearray
        :param subservice_id:
        :param time:
        :param hk_data:
        :param custom_hk_handling:  Can be used if a custom HK format is used which does not
                                    use a 8 byte structure ID (SID).
        :param minimum_reply_size:
        :param minimum_structure_report_header_size:
        """
        Service3Base.__init__(self, object_id=0, custom_hk_handling=custom_hk_handling)
        source_data = bytearray()
        source_data.extend(struct.pack("!I", self.object_id.id))
        source_data.extend(struct.pack("!I", self.set_id))
        if subservice_id == 25 or subservice_id == 26:
            source_data.extend(hk_data)
        pus_tm = PusTelemetry(
            service=3,
            subservice=subservice_id,
            time=time,
            ssc=ssc,
            source_data=source_data,
            apid=apid,
            packet_version=packet_version,
            pus_version=pus_version,
            secondary_header_flag=secondary_header_flag,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )
        PusTmBase.__init__(self, pus_tm=pus_tm)
        PusTmInfoBase.__init__(self, pus_tm=pus_tm)
        self.__init_without_base(
            instance=self,
            custom_hk_handling=custom_hk_handling,
            minimum_reply_size=minimum_reply_size,
            minimum_structure_report_header_size=minimum_structure_report_header_size,
            check_tm_data_size=False,
        )

    @staticmethod
    def __init_without_base(
        instance: Service3TM,
        custom_hk_handling: bool,
        check_tm_data_size: bool,
        minimum_reply_size: int = DEFAULT_MINIMAL_PACKET_SIZE,
        minimum_structure_report_header_size: int = STRUCTURE_REPORT_FIXED_HEADER_SIZE,
    ):
        instance.custom_hk_handling = custom_hk_handling
        if instance.custom_hk_handling:
            return
        tm_data = instance.tm_data
        if len(tm_data) < 8:
            LOGGER.warning(
                f"Invalid Service 3 packet, is too short. Detected TM data length: {len(tm_data)}"
            )
            raise ValueError
        instance.min_hk_reply_size = minimum_reply_size
        instance.hk_structure_report_header_size = minimum_structure_report_header_size
        instance.object_id.id = tm_data[0:4]
        instanceset_id = struct.unpack("!I", tm_data[4:8])[0]
        if instance.subservice == 25 or instance.subservice == 26:
            if len(tm_data) > 8:
                instance._param_length = len(tm_data[8:])
        instance.set_packet_info("Housekeeping Packet")

    @classmethod
    def __empty(cls) -> Service3TM:
        return cls(
            subservice_id=-1,
            time=CdsShortTimestamp.init_from_current_time(),
            hk_data=bytearray(),
        )

    @classmethod
    def unpack(
        cls,
        raw_telemetry: bytearray,
        custom_hk_handling: bool,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
    ) -> Service3TM:
        service_3_tm = cls.__empty()
        service_3_tm.pus_tm = PusTelemetry.unpack(
            raw_telemetry=raw_telemetry, pus_version=pus_version
        )
        service_3_tm.__init_without_base(
            instance=service_3_tm,
            custom_hk_handling=custom_hk_handling,
            check_tm_data_size=True,
        )
        return service_3_tm

    @abstractmethod
    def append_telemetry_content(self, content_list: list):
        super().append_telemetry_content(content_list=content_list)
        content_list.append(self.object_id.as_string)
        content_list.append(hex(self.set_id))
        content_list.append(int(self._param_length))

    @abstractmethod
    def append_telemetry_column_headers(self, header_list: list):
        super().append_telemetry_column_headers(header_list=header_list)
        header_list.append("Object ID")
        header_list.append("Set ID")
        header_list.append("HK data size")

    def get_hk_definitions_list(self) -> Tuple[List, List]:
        tm_data = self.tm_data
        if len(tm_data) < self.hk_structure_report_header_size:
            LOGGER.warning(
                f"Service3TM: handle_filling_definition_arrays: Invalid structure report "
                f"from {self.object_id.as_string}, is shorter "
                f"than {self.hk_structure_report_header_size}"
            )
            return [], []
        definitions_header = [
            "Object ID",
            "Set ID",
            "Report Status",
            "Is valid",
            "Collection Interval (s)",
            "Number Of IDs",
        ]
        reporting_enabled = tm_data[8]
        set_valid = tm_data[9]
        collection_interval_seconds = struct.unpack("!f", tm_data[10:14])[0] / 1000.0
        num_params = tm_data[14]
        if len(tm_data) < self.hk_structure_report_header_size + num_params * 4:
            LOGGER.warning(
                f"Service3TM: handle_filling_definition_arrays: Invalid structure report "
                f"from {self.object_id.as_string}, is shorter than "
                f"{self.hk_structure_report_header_size + num_params * 4}"
            )
            return [], []

        parameters = []
        counter = 1
        for array_index in range(
            self.hk_structure_report_header_size,
            self.hk_structure_report_header_size + 4 * num_params,
            4,
        ):
            parameter = struct.unpack("!I", tm_data[array_index : array_index + 4])[0]
            definitions_header.append("Pool ID " + str(counter))
            parameters.append(str(hex(parameter)))
            counter = counter + 1
        if reporting_enabled == 1:
            status_string = "On"
        else:
            status_string = "Off"
        if set_valid:
            valid_string = "Yes"
        else:
            valid_string = "No"
        definitions_content = [
            self.object_id.as_string,
            self._set_id,
            status_string,
            valid_string,
            collection_interval_seconds,
            num_params,
        ]
        definitions_content.extend(parameters)
        return definitions_header, definitions_content



"""Contains definitions and functions related to PUS Service 3 Telecommands.
"""
import enum
import struct

from spacepackets.ecss.tc import PusTelecommand


class Srv3Subservice(enum.IntEnum):
    ENABLE_PERIODIC_HK_GEN = 5
    DISABLE_PERIODIC_HK_GEN = 6
    ENABLE_PERIODIC_DIAGNOSTICS_GEN = 7
    DISABLE_PERIODIC_DIAGNOSTICS_GEN = 8

    REPORT_HK_REPORT_STRUCTURES = 9
    REPORT_DIAG_REPORT_STRUCTURES = 11

    HK_DEFINITIONS_REPORT = 10
    DIAG_DEFINITION_REPORT = 12

    HK_REPORT = 25
    DIAGNOSTICS_REPORT = 26

    GENERATE_ONE_PARAMETER_REPORT = 27
    GENERATE_ONE_DIAGNOSTICS_REPORT = 28

    MODIFY_PARAMETER_REPORT_COLLECTION_INTERVAL = 31
    MODIFY_DIAGNOSTICS_REPORT_COLLECTION_INTERVAL = 32


def make_sid(object_id: bytearray, set_id: int) -> bytearray:
    set_id_bytearray = struct.pack(">I", set_id)
    return object_id + set_id_bytearray


def make_interval(interval_seconds: float) -> bytearray:
    return bytearray(struct.pack("!f", interval_seconds))


def generate_one_hk_command(sid: bytearray, ssc: int):
    return PusTelecommand(service=3, subservice=27, ssc=ssc, app_data=sid)


def generate_one_diag_command(sid: bytearray, ssc: int):
    return PusTelecommand(service=3, subservice=28, ssc=ssc, app_data=sid)



"""
:file:      tcpip_tcp_com_if.py
:date:      13.05.2021
:brief:     TCP communication interface
:author:    R. Mueller
"""
import socket
import time
import enum
import threading
import select
from collections import deque
from typing import Union, Optional, Tuple

from spacepackets.ccsds.spacepacket import parse_space_packets

from tmtccmd.utility.logger import get_console_logger
from tmtccmd.config.definitions import CoreModeList
from tmtccmd.com_if.com_interface_base import CommunicationInterface
from tmtccmd.tm.definitions import TelemetryListT
from tmtccmd.utility.tmtc_printer import TmTcPrinter
from tmtccmd.config.definitions import EthernetAddressT
from tmtccmd.utility.conf_util import acquire_timeout

LOGGER = get_console_logger()

TCP_RECV_WIRETAPPING_ENABLED = False
TCP_SEND_WIRETAPPING_ENABLED = False


class TcpCommunicationType(enum.Enum):
    """Parse for space packets in the TCP stream, using the space packet header"""

    SPACE_PACKETS = 0


# pylint: disable=abstract-method
# pylint: disable=arguments-differ
# pylint: disable=too-many-arguments
class TcpIpTcpComIF(CommunicationInterface):
    """Communication interface for TCP communication."""

    DEFAULT_LOCK_TIMEOUT = 0.4
    TM_LOOP_DELAY = 0.2

    def __init__(
        self,
        com_if_key: str,
        com_type: TcpCommunicationType,
        space_packet_ids: Tuple[int],
        tm_polling_freqency: float,
        tm_timeout: float,
        tc_timeout_factor: float,
        send_address: EthernetAddressT,
        max_recv_size: int,
        max_packets_stored: int = 50,
        tmtc_printer: Union[None, TmTcPrinter] = None,
        init_mode: int = CoreModeList.LISTENER_MODE,
    ):
        """Initialize a communication interface to send and receive TMTC via TCP
        :param com_if_key:
        :param com_type:                Communication Type. By default, it is assumed that
                                        space packets are sent via TCP
        :param space_packet_ids:        16 bit packet header for space packet headers. Used to
                                        detect the start of PUS packets
        :param tm_polling_freqency:     Polling frequency in seconds
        :param tm_timeout:              Timeout in seconds
        :param tmtc_printer: Printer instance, can be passed optionally to allow packet debugging
        """
        super().__init__(com_if_key=com_if_key, tmtc_printer=tmtc_printer)
        self.tm_timeout = tm_timeout
        self.com_type = com_type
        self.space_packet_ids = space_packet_ids
        self.tc_timeout_factor = tc_timeout_factor
        self.tm_polling_frequency = tm_polling_freqency
        self.send_address = send_address
        self.max_recv_size = max_recv_size
        self.max_packets_stored = max_packets_stored
        self.init_mode = init_mode

        self.__tcp_socket: Optional[socket.socket] = None
        self.__last_connection_time = 0
        self.__tm_thread_kill_signal = threading.Event()
        # Separate thread to request TM packets periodically if no TCs are being sent
        self.__tcp_conn_thread = threading.Thread(
            target=self.__tcp_tm_client, daemon=True
        )
        self.__tm_queue = deque()
        self.__analysis_queue = deque()
        # Only allow one connection to OBSW at a time for now by using this lock
        # self.__socket_lock = threading.Lock()
        self.__queue_lock = threading.Lock()

    def __del__(self):
        try:
            self.close()
        except IOError:
            LOGGER.warning("Could not close UDP communication interface!")

    def initialize(self, args: any = None) -> any:
        self.__tm_thread_kill_signal.clear()
        self.__tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__tcp_socket.connect(self.send_address)

    def open(self, args: any = None):
        self.__tcp_conn_thread.start()

    def close(self, args: any = None) -> None:
        self.__tm_thread_kill_signal.set()
        if self.__tcp_conn_thread.is_alive():
            self.__tcp_conn_thread.join(self.tm_polling_frequency)
        try:
            self.__tcp_socket.shutdown(socket.SHUT_RDWR)
        except OSError:
            LOGGER.warning("TCP socket endpoint was already closed or not connected")
        self.__tcp_socket.close()

    def send(self, data: bytearray):
        try:
            self.__tcp_socket.sendto(data, self.send_address)
        except ConnectionRefusedError:
            LOGGER.warning("TCP connection attempt failed..")

    def receive(self, poll_timeout: float = 0) -> TelemetryListT:
        tm_packet_list = []
        with acquire_timeout(
            self.__queue_lock, timeout=self.DEFAULT_LOCK_TIMEOUT
        ) as acquired:
            if not acquired:
                LOGGER.warning("Acquiring queue lock failed!")
            while self.__tm_queue:
                self.__analysis_queue.appendleft(self.__tm_queue.pop())
        # TCP is stream based, so there might be broken packets or multiple packets in one recv
        # call. We parse the space packets contained in the stream here
        if self.com_type == TcpCommunicationType.SPACE_PACKETS:
            tm_packet_list = parse_space_packets(
                analysis_queue=self.__analysis_queue, packet_ids=self.space_packet_ids
            )
        else:
            while self.__analysis_queue:
                tm_packet_list.append(self.__analysis_queue.pop())
        return tm_packet_list

    def __tcp_tm_client(self):
        while True and not self.__tm_thread_kill_signal.is_set():
            try:
                self.__receive_tm_packets()
            except ConnectionRefusedError:
                LOGGER.warning("TCP connection attempt failed..")
            time.sleep(self.TM_LOOP_DELAY)

    def __receive_tm_packets(self):
        try:
            ready = select.select([self.__tcp_socket], [], [], 0)
            if ready[0]:
                bytes_recvd = self.__tcp_socket.recv(self.max_recv_size)
                with acquire_timeout(
                    self.__queue_lock, timeout=self.DEFAULT_LOCK_TIMEOUT
                ) as acquired:
                    if not acquired:
                        LOGGER.warning("Acquiring queue lock failed!")
                    if self.__tm_queue.__len__() >= self.max_packets_stored:
                        LOGGER.warning(
                            "Number of packets in TCP queue too large. "
                            "Overwriting old packets.."
                        )
                        self.__tm_queue.pop()
                    self.__tm_queue.appendleft(bytearray(bytes_recvd))
        except ConnectionResetError:
            LOGGER.exception("ConnectionResetError. TCP server might not be up")

    def data_available(self, timeout: float = 0, parameters: any = 0) -> bool:
        if self.__tm_queue:
            return True
        else:
            return False



"""Dummy Communication Interface. Currently serves to provide an example without external hardware
"""
from spacepackets.ecss.tc import PusTelecommand
from spacepackets.ccsds.spacepacket import (
    get_space_packet_sequence_control,
    SequenceFlags,
)

from tmtccmd.com_if.com_interface_base import CommunicationInterface
from tmtccmd.tm import TelemetryListT
from tmtccmd.pus.service_1_verification import Service1TMExtended
from tmtccmd.pus.service_17_test import Srv17Subservices, Service17TMExtended
from tmtccmd.utility.logger import get_console_logger
from tmtccmd.utility.tmtc_printer import TmTcPrinter

LOGGER = get_console_logger()


class DummyComIF(CommunicationInterface):
    def __init__(self, com_if_key: str, tmtc_printer: TmTcPrinter):
        super().__init__(com_if_key=com_if_key, tmtc_printer=tmtc_printer)
        self.dummy_handler = DummyHandler()
        self.last_service = 0
        self.last_subservice = 0
        self.tc_ssc = 0
        self.tc_packet_id = 0

    def initialize(self, args: any = None) -> any:
        pass

    def open(self, args: any = None) -> None:
        pass

    def close(self, args: any = None) -> None:
        pass

    def data_available(self, timeout: float = 0, parameters: any = 0):
        if self.dummy_handler.reply_pending:
            return True
        return False

    def receive(self, parameters: any = 0) -> TelemetryListT:
        return self.dummy_handler.receive_reply_package()

    def send(self, data: bytearray):
        if data is not None:
            self.dummy_handler.pass_telecommand(data=data)


class DummyHandler:
    def __init__(self):
        self.last_telecommand = None
        self.next_telemetry_package = []
        self.last_tc_ssc = 0
        self.last_tc_packet_id = 0
        self.current_ssc = 0
        self.reply_pending = False

    def pass_telecommand(self, data: bytearray):
        # TODO: Need TC deserializer for cleaner implementation
        self.last_telecommand = data
        self.last_tc_ssc = ((data[2] << 8) | data[3]) & 0x3FFF
        self.last_service = data[7]
        self.last_subservice = data[8]
        self.tc_packet_id = data[0] << 8 | data[1]
        self.reply_pending = True
        self.generate_reply_package()

    def generate_reply_package(self):
        """
        Generate the replies. This function will perform the following steps:
         - Generate an object representation of the telemetry to be generated based on service and subservice
         - Generate the raw bytearray of the telemetry
         - Generate the object representation which would otherwise be generated from the raw bytearray received
           from an external source
        """
        if self.last_service == 17:
            if self.last_subservice == 1:
                tc_psc = get_space_packet_sequence_control(
                    sequence_flags=SequenceFlags.UNSEGMENTED,
                    source_sequence_count=self.last_tc_ssc,
                )
                tm_packer = Service1TMExtended(
                    subservice=1,
                    ssc=self.current_ssc,
                    tc_packet_id=self.last_tc_packet_id,
                    tc_psc=tc_psc,
                )

                self.current_ssc += 1
                tm_packet_raw = tm_packer.pack()
                self.next_telemetry_package.append(tm_packet_raw)
                tm_packer = Service1TMExtended(
                    subservice=7,
                    ssc=self.current_ssc,
                    tc_packet_id=self.last_tc_packet_id,
                    tc_psc=tc_psc,
                )
                tm_packet_raw = tm_packer.pack()
                self.next_telemetry_package.append(tm_packet_raw)
                self.current_ssc += 1

                tm_packer = Service17TMExtended(subservice=Srv17Subservices.PING_REPLY)
                tm_packet_raw = tm_packer.pack()
                self.next_telemetry_package.append(tm_packet_raw)
                self.current_ssc += 1

    def receive_reply_package(self) -> TelemetryListT:
        if self.reply_pending:
            return_list = self.next_telemetry_package.copy()
            self.next_telemetry_package.clear()
            self.reply_pending = False
            return return_list
        else:
            return []



from typing import Dict


INVALID_ID = bytes([0xFF, 0xFF, 0xFF, 0xFF])


def get_core_object_ids() -> Dict[bytes, list]:
    """
    These are the object IDs for the tmtccmd core. The core will usually take care of
    inserting these into the object manager during the program initialization.
    :return: Dictionary of the core object IDs
    """
    object_id_dict = {
        INVALID_ID: ["Invalid ID"],
    }
    return object_id_dict



# -*- coding: utf-8 -*-
"""
:file:      tmtcc_com_interface_base.py
:data:      01.11.2019
:details:
Generic Communication Interface. Defines the syntax of the communication functions.
Abstract methods must be implemented by child class (e.g. Ethernet Com IF)

:author:     R. Mueller
"""
from abc import abstractmethod

from tmtccmd.tm.definitions import TelemetryListT
from tmtccmd.utility.tmtc_printer import TmTcPrinter


# pylint: disable=useless-return
# pylint: disable=no-self-use
# pylint: disable=unused-argument
class CommunicationInterface:
    """Generic form of a communication interface to separate communication logic from
    the underlying interface.
    """

    def __init__(self, tmtc_printer: TmTcPrinter, com_if_key: str):
        self.tmtc_printer = tmtc_printer
        self.valid = True
        self.com_if_key = com_if_key

    def get_id(self) -> str:
        return self.com_if_key

    @abstractmethod
    def initialize(self, args: any = None) -> any:
        """Perform initializations step which can not be done in constructor or which require
        returnvalues.
        """

    @abstractmethod
    def open(self, args: any = None) -> None:
        """Opens the communication interface to allow communication.
        @return:
        """

    @abstractmethod
    def close(self, args: any = None) -> None:
        """Closes the ComIF and releases any held resources (for example a Communication Port).
        :return:
        """

    @abstractmethod
    def send(self, data: bytearray):
        """Send raw data"""

    @abstractmethod
    def receive(self, parameters: any = 0) -> TelemetryListT:
        """Returns a list of received packets. The child class can use a separate thread to poll for
        the packets or use some other mechanism and container like a deque to store packets
        to be returned here.
        :param parameters:
        :return:
        """
        packet_list = []
        return packet_list

    @abstractmethod
    def data_available(self, timeout: float, parameters: any) -> int:
        """Check whether TM data is available
        :param parameters: Can be an arbitrary parameter like a timeout
        :return: 0 if no data is available, number of bytes or anything > 0 otherwise.
        """



"""
@brief      Core components for mode commanding (custom PUS service)
"""
import enum
import struct


class Modes(enum.IntEnum):
    OFF = (0,)
    ON = (1,)
    NORMAL = (2,)
    RAW = 3


def pack_mode_data(object_id: bytearray, mode: Modes, submode: int) -> bytearray:
    """Mode 0: Off, Mode 1: Mode On, Mode 2: Mode Normal, Mode 3: Mode Raw"""
    # Normal mode
    mode_packed = struct.pack("!I", mode)
    # Submode default
    submode_byte = struct.pack("B", submode)
    mode_data = object_id + mode_packed + submode_byte
    return mode_data



from typing import Tuple, Union, Deque
from tmtccmd.config.definitions import QueueCommands
from spacepackets.ecss.tc import PusTelecommand

TcAuxiliaryTupleT = Tuple[QueueCommands, any]
PusTcTupleT = Tuple[bytearray, Union[None, PusTelecommand]]
TcQueueEntryT = Union[TcAuxiliaryTupleT, PusTcTupleT]
TcQueueT = Deque[TcQueueEntryT]



VERSION_NAME = "tmtccmd"
VERSION_MAJOR = 1
VERSION_MINOR = 10
VERSION_REVISION = 1

# I think this needs to be in string representation to be parsed so we can't
# use a formatted string here.
__version__ = "1.10.1"



import json

from tmtccmd.utility.logger import get_console_logger
from tmtccmd.utility.json_handler import check_json_file, JsonKeyNames
from tmtccmd.config.definitions import ComIFDictT

LOGGER = get_console_logger()


def determine_com_if(com_if_dict: ComIFDictT, json_cfg_path: str) -> str:
    do_prompt_com_if = False
    if not check_json_file(json_cfg_path=json_cfg_path):
        do_prompt_com_if = True
    if not do_prompt_com_if:
        with open(json_cfg_path, "r") as read:
            com_if_string = ""
            try:
                load_data = json.load(read)
                com_if_string = load_data[JsonKeyNames.COM_IF.value]
            except KeyError:
                do_prompt_com_if = True
            com_if_string = str(com_if_string)
    if do_prompt_com_if:
        com_if_string = prompt_com_if(com_if_dict=com_if_dict)
        save_to_json = input(
            "Do you want to store the communication interface? (y/n): "
        )
        if save_to_json.lower() in ["y", "yes", "1"]:
            store_com_if_json(com_if_string=com_if_string, json_cfg_path=json_cfg_path)
    return com_if_string


def prompt_com_if(com_if_dict: dict) -> str:
    while True:
        com_if_list = []
        for index, com_if_value in enumerate(com_if_dict.items()):
            print(f"{index}: {com_if_value}")
            com_if_list.append(com_if_value)
        com_if_key = input("Please enter the desired communication interface by key: ")
        if not com_if_key.isdigit():
            print("Key is not a digit, try again")
            continue
        com_if_key = int(com_if_key)
        if com_if_key >= len(com_if_list):
            print("Key invalid, try again.")
            continue
        com_if_string = com_if_list[com_if_key][0]
        break
    return com_if_string


def store_com_if_json(com_if_string: str, json_cfg_path: str):
    with open(json_cfg_path, "r+") as file:
        data = json.load(file)
        data[JsonKeyNames.COM_IF.value] = com_if_string
        file.seek(0)
        json.dump(data, file, indent=4)
    LOGGER.info(f"Communication interface was stored in the JSON file {json_cfg_path}")
    LOGGER.info(
        "Delete this file or edit it manually to edit the communication interface"
    )



import enum
import struct

from tmtccmd.tc.definitions import PusTelecommand
from spacepackets.ecss.conf import get_default_tc_apid


class Srv8Subservices(enum.IntEnum):
    FUNC_CMD = (128,)
    DATA_REPLY = 130


def generate_action_command(
    object_id: bytes,
    action_id: int,
    app_data: bytes = bytes(),
    ssc: int = 0,
    apid: int = -1,
) -> PusTelecommand:
    if apid == -1:
        apid = get_default_tc_apid()
    data_to_pack = bytearray(object_id)
    data_to_pack += make_action_id(action_id) + app_data
    return PusTelecommand(
        service=8,
        subservice=Srv8Subservices.FUNC_CMD,
        ssc=ssc,
        app_data=data_to_pack,
        apid=apid,
    )


def make_action_id(action_id: int) -> bytearray:
    return bytearray(struct.pack("!I", action_id))



from typing import Tuple, List
import enum

from tmtccmd.pus import ObjectId


class HkContentType(enum.Enum):
    HK = enum.auto()
    DEFINITIONS = enum.auto()


class Service3Base:
    """Base class. The TMTC core provides a Service 3 implementation which is intended to be used
    with the FSFW. However, users can define an own Service 3 implementation.

    The TMTC printer utility uses the fields defined in this base class to perform prinouts so
    if a custom class is defined, the user should implement this class and fill the fields
    in the TM handling hook if printout of the HK field and validity checking is desired.
    """

    def __init__(self, object_id: int, custom_hk_handling: bool = False):
        self._object_id = ObjectId(object_id=object_id)
        self._set_id = 0
        self._param_length = 0
        self._custom_hk_handling = custom_hk_handling

    @property
    def object_id(self) -> ObjectId:
        return self._object_id

    @property
    def set_id(self) -> int:
        return self._set_id

    @property
    def has_custom_hk_handling(self) -> bool:
        return self._custom_hk_handling

    @has_custom_hk_handling.setter
    def has_custom_hk_handling(self, has_custom_hk_handling: bool):
        self._custom_hk_handling = has_custom_hk_handling

    @property
    def hk_definitions_list(self) -> Tuple[List, List]:
        """Can be implemented by a child class to print definitions lists. The first list
        should contain a header with parameter names, and the second list shall contain the
        corresponding set IDs"""
        return [], []



"""
Program: obsw_module_test.py
Date: 01.11.2019
Description: All functions related to TmTc Sending and Receiving, used by UDP client

Manual:
Set up the UDP client as specified in the header comment and use the unit testing mode

A separate thread is used to listen for replies and send a new telecommand
if the first reply has not been received.

@author: R. Mueller
"""
import time

from tmtccmd.com_if.com_interface_base import CommunicationInterface
from tmtccmd.config.definitions import QueueCommands, CoreGlobalIds
from tmtccmd.utility.tmtc_printer import TmTcPrinter
from tmtccmd.utility.logger import get_console_logger

from tmtccmd.ccsds.handler import CcsdsTmHandler
from tmtccmd.sendreceive.tm_listener import TmListener
from tmtccmd.tc.definitions import TcQueueEntryT
from tmtccmd.core.globals_manager import get_global

LOGGER = get_console_logger()


# pylint: disable=too-many-instance-attributes
class CommandSenderReceiver:
    """
    This is the generic CommandSenderReceiver object. All TMTC objects inherit this object,
    for example specific implementations (e.g. SingleCommandSenderReceiver)
    """

    def __init__(
        self,
        com_if: CommunicationInterface,
        tmtc_printer: TmTcPrinter,
        tm_listener: TmListener,
        tm_handler: CcsdsTmHandler,
        apid: int,
    ):

        """
        :param com_if: CommunicationInterface object. Instantiate the desired one
        and pass it here
        :param tmtc_printer: TmTcPrinter object. Instantiate it and pass it here.
        """
        self._tm_timeout = get_global(CoreGlobalIds.TM_TIMEOUT)
        self._tm_handler = tm_handler
        self._tc_send_timeout_factor = get_global(CoreGlobalIds.TC_SEND_TIMEOUT_FACTOR)
        self._apid = apid

        if isinstance(com_if, CommunicationInterface):
            self._com_if = com_if
        else:
            LOGGER.error("CommandSenderReceiver: Invalid communication interface!")
            raise TypeError("CommandSenderReceiver: Invalid communication interface!")

        if isinstance(tmtc_printer, TmTcPrinter):
            self._tmtc_printer = tmtc_printer
        else:
            LOGGER.error("CommandSenderReceiver: Invalid TMTC printer!")
            raise TypeError("CommandSenderReceiver: Invalid TMTC printer!")

        if isinstance(tm_listener, TmListener):
            self._tm_listener = tm_listener
        else:
            LOGGER.error("CommandSenderReceiver: Invalid TM listener!")
            raise TypeError("Invalid TM Listener!")

        self._start_time = 0
        self._elapsed_time = 0
        self._timeout_counter = 0

        # needed to store last actual TC packet from queue
        self._last_tc = bytearray()
        self._last_tc_obj = None

        # this flag can be used to notify when the operation is finished
        self._operation_pending = False
        # This flag can be used to notify when a reply was received.
        self._reply_received = False

        self._wait_period = 0
        self._wait_start = 0

    def set_tm_timeout(self, tm_timeout: float = -1):
        """
        Set the TM timeout. Usually, the global value set by the args parser is set,
        but the TM timeout can be reset (e.g. for slower architectures)
        :param tm_timeout: New TM timeout value as a float value in seconds
        :return:
        """
        if tm_timeout == -1:
            tm_timeout = get_global(CoreGlobalIds.TM_TIMEOUT)
        self._tm_timeout = tm_timeout

    def set_tc_send_timeout_factor(self, new_factor: float = -1):
        """
        Set the TC resend timeout factor. After self._tm_timeout * new_factor seconds,
        a telecommand will be resent again.
        :param new_factor: Factor as a float number
        :return:
        """
        if new_factor == -1:
            new_factor = get_global(CoreGlobalIds.TC_SEND_TIMEOUT_FACTOR)
        self._tc_send_timeout_factor = new_factor

    def _check_for_first_reply(self) -> None:
        """
        Checks for replies. If no reply is received, send telecommand again in checkForTimeout()
        :return: None
        """
        if self._tm_listener.reply_event():
            self._reply_received = True
            self._operation_pending = False
            self._tm_listener.clear_reply_event()
        else:
            self._check_for_timeout()

    def wait_period_ongoing(
        self,
        sleep_rest_of_wait_period: bool = False,
        set_reply_rcvd_to_true: bool = True,
    ):
        if sleep_rest_of_wait_period:
            # wait rest of wait time
            sleep_time = self._wait_start + self._wait_period - time.time()
            if sleep_time > 0:
                time.sleep(sleep_time)
            LOGGER.info("Wait period over.")
            return False
        # If wait period was specified, we need to wait before checking the next queue entry.
        if self._wait_period > 0:
            if time.time() - self._wait_start < self._wait_period:
                if set_reply_rcvd_to_true:
                    self._reply_received = True
                return True
            else:
                LOGGER.info("Wait period over.")
                self._wait_period = 0
                return False
        else:
            return False

    @staticmethod
    def check_queue_entry_static(tc_queue_entry: TcQueueEntryT) -> bool:
        """Static method to check whether a queue entry is a valid telecommand"""
        queue_entry_first, queue_entry_second = tc_queue_entry
        if isinstance(queue_entry_first, str):
            LOGGER.warning("Invalid telecommand. Queue entry is a string!")
            return False
        if isinstance(queue_entry_first, QueueCommands):
            return False
        elif isinstance(queue_entry_first, bytearray):
            return True
        else:
            return False

    def check_queue_entry(self, tc_queue_entry: TcQueueEntryT) -> bool:
        """
        Checks whether the entry in the pus_tc queue is a telecommand.
        The last telecommand and respective information are stored in _last_tc
        and _last_tc_info
        :param tc_queue_entry:
        :return: True if queue entry is telecommand, False if it is not
        """
        queue_entry_first, queue_entry_second = tc_queue_entry
        queue_entry_is_telecommand = False

        if isinstance(queue_entry_first, str):
            LOGGER.warning("Invalid telecommand. Queue entry is a string!")
            return queue_entry_is_telecommand

        if queue_entry_first == QueueCommands.WAIT:
            wait_time = queue_entry_second
            self._tm_timeout = self._tm_timeout + wait_time
            self._wait_period = wait_time
            print()
            print_string = "Waiting for " + str(self._wait_period) + " seconds."
            LOGGER.info(print_string)
            self._wait_start = time.time()
        # printout optimized for LOGGER and debugging
        elif queue_entry_first == QueueCommands.PRINT:
            print_string = queue_entry_second
            print()
            self._tmtc_printer.print_string(print_string, True)
        elif queue_entry_first == QueueCommands.RAW_PRINT:
            print_string = queue_entry_second
            self._tmtc_printer.print_string(print_string, False)
        elif queue_entry_first == QueueCommands.EXPORT_LOG:
            export_name = queue_entry_second
            self._tmtc_printer.add_print_buffer_to_buffer_list()
            self._tmtc_printer.print_to_file(export_name, True)
        elif queue_entry_first == QueueCommands.SET_TIMEOUT:
            self._tm_timeout = queue_entry_second
        else:
            self._last_tc, self._last_tc_obj = (queue_entry_first, queue_entry_second)
            return True
        return queue_entry_is_telecommand

    def _check_for_timeout(self, last_timeout: bool = True):
        """
        Checks whether a timeout after sending a telecommand has occured and sends telecommand
        again. If resending reached certain counter, exit the program.
        :return:
        """

        if self._start_time == 0:
            self._start_time = time.time()
        if self._timeout_counter == 5:
            LOGGER.info("CommandSenderReceiver: No response from command !")
            self._operation_pending = False
        if self._start_time != 0:
            self._elapsed_time = time.time() - self._start_time
        if self._elapsed_time >= self._tm_timeout * self._tc_send_timeout_factor:
            from tmtccmd.core.globals_manager import get_global

            if get_global(CoreGlobalIds.RESEND_TC):
                LOGGER.info("CommandSenderReceiver: Timeout, sending TC again !")
                self._com_if.send(self._last_tc)
                self._timeout_counter = self._timeout_counter + 1
                self._start_time = time.time()
            else:
                # todo: we could also stop sending and clear the TC queue
                self._reply_received = True
        time.sleep(0.5)

    # TODO: Move to TMTC printer to decouple this module?
    # def print_tm_queue(self, tm_queue: TelemetryQueueT):
    #    while tm_queue:
    #        try:
    #            tm_packet_list = tm_queue.pop()
    #            for tm_packet in tm_packet_list:
    #                telemetry = PusTelemetry(tm_packet)
    #                self._tmtc_printer.print_telemetry()
    #        except AttributeError as e:
    #            LOGGER.exception("CommandSenderReceiver Exception: Invalid queue entry. Traceback:", e)



from unittest.mock import MagicMock

from tmtccmd.core.backend import TmTcHandler
from tmtccmd.ccsds.handler import CcsdsTmHandler
from tmtccmd.config.com_if import create_communication_interface_default
from tmtccmd.sendreceive.tm_listener import TmListener
from tmtccmd.utility.tmtc_printer import TmTcPrinter, DisplayMode
from tmtccmd.config.definitions import CoreComInterfaces, CoreModeList
from tmtccmd.core.frontend_base import FrontendBase


def create_backend_mock(tm_handler: CcsdsTmHandler) -> TmTcHandler:
    tmtc_printer = TmTcPrinter(
        display_mode=DisplayMode.LONG, do_print_to_file=False, print_tc=True
    )
    com_if = create_communication_interface_default(
        com_if_key=CoreComInterfaces.DUMMY.value,
        json_cfg_path="tmtc_config.json",
        tmtc_printer=tmtc_printer,
    )
    tm_listener = TmListener(com_if=com_if, tm_timeout=3.0, tc_timeout_factor=3.0)
    # The global variables are set by the argument parser.
    tmtc_backend = TmTcHandler(
        com_if=com_if,
        tmtc_printer=tmtc_printer,
        tm_listener=tm_listener,
        init_mode=CoreModeList.IDLE,
        init_service=17,
        init_opcode="0",
        tm_handler=tm_handler,
    )
    tmtc_backend.start_listener = MagicMock(return_value=0)
    tmtc_backend.initialize = MagicMock(return_value=0)
    return tmtc_backend


def create_frontend_mock() -> FrontendBase:
    from tmtccmd.core.frontend_base import FrontendBase

    tmtc_frontend = FrontendBase()
    tmtc_frontend.start = MagicMock(return_value=0)
    return tmtc_frontend



"""Base class for Service 200 mode commanding reply handling.
"""
from __future__ import annotations
import struct
from spacepackets.ecss.tm import CdsShortTimestamp, PusVersion, PusTelemetry

from tmtccmd.pus.definitions import CustomPusServices
from tmtccmd.tm.base import PusTmInfoBase, PusTmBase


class Service200TM(PusTmBase, PusTmInfoBase):
    def __init__(
        self,
        subservice_id: int,
        object_id: bytearray,
        return_value: int = 0,
        mode: int = 0,
        submode: int = 0,
        time: CdsShortTimestamp = None,
        ssc: int = 0,
        source_data: bytearray = bytearray([]),
        apid: int = -1,
        packet_version: int = 0b000,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
        pus_tm_version: int = 0b0001,
        secondary_header_flag: bool = True,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        source_data = bytearray()
        source_data.extend(object_id)
        if subservice_id == 7:
            source_data.extend(struct.pack("!H", return_value))
        elif subservice_id == 6 or subservice_id == 8:
            source_data.extend(struct.pack("!I", mode))
            source_data.append(submode)
        pus_tm = PusTelemetry(
            service=CustomPusServices.SERVICE_200_MODE,
            subservice=subservice_id,
            time=time,
            ssc=ssc,
            source_data=source_data,
            apid=apid,
            packet_version=packet_version,
            pus_version=pus_version,
            secondary_header_flag=secondary_header_flag,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )
        PusTmBase.__init__(self, pus_tm=pus_tm)
        PusTmInfoBase.__init__(self, pus_tm=pus_tm)
        self.set_packet_info("Mode Reply")

        self.is_cant_reach_mode_reply = False
        self.is_mode_reply = False
        self.object_id = object_id
        self.mode = mode
        self.submode = submode
        self.return_value = return_value
        self.__init_without_base(instance=self)

    @staticmethod
    def __init_without_base(instance: Service200TM):
        tm_data = instance.tm_data
        instance.object_id = tm_data[0:4]
        if instance.subservice == 7:
            instance.append_packet_info(": Can't reach mode")
            instance.is_cant_reach_mode_reply = True
            instance.return_value = tm_data[4] << 8 | tm_data[5]
        elif instance.subservice == 6 or instance.subservice == 8:
            instance.is_mode_reply = True
            if instance.subservice == 8:
                instance.append_packet_info(": Wrong Mode")
            elif instance.subservice == 6:
                instance.append_packet_info(": Mode reached")
            instance.mode = struct.unpack("!I", tm_data[4:8])[0]
            instance.submode = tm_data[8]

    @classmethod
    def __empty(cls) -> Service200TM:
        return cls(subservice_id=0, object_id=bytearray(4))

    @classmethod
    def unpack(
        cls,
        raw_telemetry: bytearray,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
    ) -> Service200TM:
        service_200_tm = cls.__empty()
        service_200_tm.pus_tm = PusTelemetry.unpack(
            raw_telemetry=raw_telemetry, pus_version=pus_version
        )
        service_200_tm.__init_without_base(instance=service_200_tm)
        return service_200_tm

    def append_telemetry_content(self, content_list: list):
        super().append_telemetry_content(content_list=content_list)
        object_id_num = struct.unpack("!I", self.object_id)[0]
        content_list.append(f"0x{object_id_num:08x}")
        if self.is_cant_reach_mode_reply:
            content_list.append(hex(self.return_value))
        elif self.is_mode_reply:
            if self.mode == 1:
                mode_string = "ON"
            elif self.mode == 2:
                mode_string = "NORMAL"
            elif self.mode == 3:
                mode_string = "RAW"
            elif self.mode == 0:
                mode_string = "OFF"
            else:
                mode_string = f"UNKNOWN ({self.mode})"
            content_list.append(mode_string)
            content_list.append(str(self.submode))

    def append_telemetry_column_headers(self, header_list: list):
        super().append_telemetry_column_headers(header_list=header_list)
        header_list.append("Object ID")
        if self.is_cant_reach_mode_reply:
            header_list.append("Return Value")
        elif self.is_mode_reply:
            header_list.append("Mode")
            header_list.append("Submode")



# -*- coding: utf-8 -*-
"""Contains classes and functions to deserialize PUS Service 5 Telemetry
"""
from __future__ import annotations
from abc import abstractmethod
import struct

from spacepackets.ecss.definitions import PusServices
from spacepackets.ecss.tm import CdsShortTimestamp, PusVersion
from tmtccmd.tm.base import PusTmInfoBase, PusTmBase, PusTelemetry
from tmtccmd.pus.service_5_event import Srv5Subservices, Srv5Severity
from tmtccmd.pus.obj_id import ObjectId
from tmtccmd.utility.logger import get_console_logger


LOGGER = get_console_logger()


class Service5TM(PusTmBase, PusTmInfoBase):
    def __init__(
        self,
        subservice: Srv5Subservices,
        event_id: int,
        object_id: bytearray,
        param_1: int,
        param_2: int,
        time: CdsShortTimestamp = None,
        ssc: int = 0,
        apid: int = -1,
        packet_version: int = 0b000,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
        secondary_header_flag: bool = True,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        """Create a Service 5 telemetry instance.
        Use the unpack function to create an instance from a raw bytestream instead.
        :param subservice: Subservice ID
        :param time: CDS Short Timecode
        :param object_id: 4 byte object ID
        :raises ValueError: Invalid input arguments
        """
        self._object_id = ObjectId.from_bytes(obj_id_as_bytes=object_id)
        self._event_id = event_id
        self._param_1 = param_1
        self._param_2 = param_2
        source_data = bytearray()
        source_data.extend(struct.pack("!H", self._event_id))
        if len(object_id) != 4:
            LOGGER.warning("Object ID must be a bytrarray with length 4")
            raise ValueError
        source_data.extend(object_id)
        source_data.extend(struct.pack("!I", self._param_1))
        source_data.extend(struct.pack("!I", self._param_2))
        pus_tm = PusTelemetry(
            service=PusServices.SERVICE_5_EVENT,
            subservice=subservice,
            time=time,
            ssc=ssc,
            source_data=source_data,
            apid=apid,
            packet_version=packet_version,
            pus_version=pus_version,
            secondary_header_flag=secondary_header_flag,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )
        PusTmBase.__init__(self, pus_tm=pus_tm)
        PusTmInfoBase.__init__(self, pus_tm=pus_tm)
        self.__init_without_base(instance=self, set_attrs_from_tm_data=False)

    @classmethod
    def __empty(cls) -> Service5TM:
        return cls(
            subservice=Srv5Subservices.INFO_EVENT,
            event_id=0,
            object_id=bytearray(4),
            param_1=0,
            param_2=0,
        )

    @classmethod
    def unpack(
        cls,
        raw_telemetry: bytearray,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
    ) -> Service5TM:
        service_5_tm = cls.__empty()
        service_5_tm.pus_tm = PusTelemetry.unpack(
            raw_telemetry=raw_telemetry, pus_version=pus_version
        )
        service_5_tm.__init_without_base(instance=service_5_tm)
        return service_5_tm

    @abstractmethod
    def append_telemetry_content(self, content_list: list):
        super().append_telemetry_content(content_list=content_list)
        content_list.append(str(self._event_id))
        content_list.append(self._object_id.as_string)
        content_list.append(str(hex(self._param_1)) + ", " + str(self._param_1))
        content_list.append(str(hex(self._param_2)) + ", " + str(self._param_2))

    @abstractmethod
    def append_telemetry_column_headers(self, header_list: list):
        super().append_telemetry_column_headers(header_list=header_list)
        header_list.append("Event ID")
        header_list.append("Reporter ID")
        header_list.append("Parameter 1")
        header_list.append("Parameter 2")

    @property
    def reporter_id(self) -> int:
        return self._object_id.id

    @property
    def reporter_id_as_bytes(self) -> bytes:
        return self._object_id.as_bytes

    @property
    def event_id(self):
        return self._event_id

    @property
    def param_1(self):
        return self._param_1

    @property
    def param_2(self):
        return self._param_2

    @staticmethod
    def __init_without_base(instance: Service5TM, set_attrs_from_tm_data: bool = False):
        if instance.service != 5:
            LOGGER.warning("This packet is not an event service packet!")
        instance.set_packet_info("Event")
        if instance.subservice == Srv5Subservices.INFO_EVENT:
            instance.append_packet_info(" Info")
        elif instance.subservice == Srv5Subservices.LOW_SEVERITY_EVENT:
            instance.append_packet_info(" Error Low Severity")
        elif instance.subservice == Srv5Subservices.MEDIUM_SEVERITY_EVENT:
            instance.append_packet_info(" Error Med Severity")
        elif instance.subservice == Srv5Subservices.HIGH_SEVERITY_EVENT:
            instance.append_packet_info(" Error High Severity")
        tm_data = instance.tm_data
        if len(tm_data) < 14:
            LOGGER.warning(
                f"Length of TM data field {len(tm_data)} shorter than expected 14 bytes"
            )
            raise ValueError
        if set_attrs_from_tm_data:
            instance._event_id = struct.unpack('>H', tm_data[0:2])[0]
            instance._object_id.from_bytes(tm_data[2:6])
            instance._param_1 = struct.unpack('>I', tm_data[6:10])[0]
            instance._param_2 = struct.unpack('>I', tm_data[10:14])[0]



import signal
from tmtccmd.com_if.com_interface_base import CommunicationInterface
from tmtccmd.core.backend import TmTcHandler
from tmtccmd.utility.logger import get_console_logger

LOGGER = get_console_logger()


def keyboard_interrupt_handler(
    tmtc_backend: TmTcHandler, com_interface: CommunicationInterface
):
    tmtc_backend.close_listener(join=True, join_timeout_seconds=1.0)
    LOGGER.info("Closing TMTC client")


class GracefulKiller:
    kill_now = False

    def __init__(self):
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def exit_gracefully(self):
        self.kill_now = True
        print("I was killed")



"""Used to send multiple TCs as bursts and listen for replies simultaneously. Used by Module Tester
"""
import sys
import time
from typing import Union, Deque
from collections import deque

from tmtccmd.config.definitions import CoreGlobalIds
from tmtccmd.sendreceive.sequential_sender_receiver import (
    SequentialCommandSenderReceiver,
)
from tmtccmd.com_if.com_interface_base import CommunicationInterface
from tmtccmd.utility.tmtc_printer import TmTcPrinter
from tmtccmd.core.globals_manager import get_global
from tmtccmd.sendreceive.tm_listener import TmListener
from tmtccmd.utility.tmtc_printer import get_console_logger


LOGGER = get_console_logger()


class MultipleCommandSenderReceiver(SequentialCommandSenderReceiver):
    """Difference to seqential sender: This class can send TCs in bursts.
    Wait intervals can be specified with wait time between the send bursts.
    This is generally done in the separate test classes in UnitTest
    """

    def __init__(
        self,
        com_if: CommunicationInterface,
        tmtc_printer: TmTcPrinter,
        tc_queue: Deque,
        tm_listener: TmListener,
        wait_intervals: list,
        wait_time: Union[float, list],
        print_tm: bool,
    ):
        """TCs are sent in burst when applicable. Wait intervals can be specified by supplying
        respective arguments

        :param com_if:
        :param tmtc_printer:
        :param tc_queue:
        :param wait_intervals: List of pause intervals. For example [1,3] means that a wait_time
            is applied after
        sendinf the first and the third telecommand
        :param wait_time: List of wait times or uniform wait time as float
        :param print_tm:
        """
        super().__init__(
            com_if=com_if,
            tmtc_printer=tmtc_printer,
            tm_listener=tm_listener,
            tc_queue=tc_queue,
        )
        self.waitIntervals = wait_intervals
        self.waitTime = wait_time
        self.printTm = print_tm
        self.tm_packet_queue = deque()
        self.tc_info_queue = deque()
        self.pusPacketInfo = []
        self.pusPacket = []
        self.waitCounter = 0

    def send_tc_queue_and_return_info(self):
        try:
            self._tm_listener.set_listener_mode(TmListener.ListenerModes.MANUAL)
            self._tm_listener.event_mode_change.set()
            time.sleep(0.1)
            # TC info queue is set in this function
            self.__send_all_queue()
            time.sleep(self._tm_timeout / 1.4)
            # Get a copy of the queue, otherwise we will lose the data.
            tm_packet_queue_list = self._tm_listener.retrieve_tm_packet_queue().copy()
            if get_global(CoreGlobalIds.PRINT_TM):
                self.print_tm_queue(self._tm_listener.retrieve_tm_packet_queue())
            self._tm_listener.clear_tm_packet_queue()
            if get_global(CoreGlobalIds.PRINT_TO_FILE):
                self._tmtc_printer.print_to_file()
            return self.tc_info_queue, tm_packet_queue_list
        except (KeyboardInterrupt, SystemExit):
            LOGGER.info("Keyboard Interrupt.")
            sys.exit()

    def __handle_tc_resending(self):
        while not self.__all_replies_received:
            if self._tc_queue.__len__ == 0:
                if self._start_time == 0:
                    self._start_time = time.time()
            self._check_for_timeout()

    def __send_all_queue(self):
        while not self._tc_queue.__len__() == 0:
            self.__send_and_print_tc()

    def __send_and_print_tc(self):
        tc_queue_tuple = self._tc_queue.pop()
        if self.check_queue_entry(tc_queue_tuple):
            pus_packet, pus_packet_obj = tc_queue_tuple
            self.tc_info_queue.append(pus_packet_obj)
            self._com_if.send_telecommand(pus_packet, pus_packet_obj)
            self.__handle_waiting()

    def __handle_waiting(self):
        self.waitCounter = self.waitCounter + 1
        if self.waitCounter in self.waitIntervals:
            if isinstance(self.waitTime, list):
                time.sleep(self.waitTime[self.waitIntervals.index(self.waitCounter)])
            else:
                time.sleep(self.waitTime)
        if self.waitTime == 0:
            # To prevent thread starvation
            time.sleep(0.1)

    def __retrieve_listener_tm_packet_queue(self):
        if self._tm_listener.reply_event():
            return self._tm_listener.retrieve_tm_packet_queue()
        else:
            LOGGER.error(
                "Multiple Command SenderReceiver: Configuration error, "
                "reply event not set in TM listener"
            )

    def __clear_listener_tm_info_queue(self):
        self._tm_listener.clear_tm_packet_queue()



from unittest import TestCase
from tmtccmd.core.object_id_manager import insert_object_id, get_object_id_info

TEST_ID_0 = bytes([0x00, 0x01, 0x02, 0x03])


class TestObjIdManager(TestCase):
    def test_obj_id_manager(self):
        insert_object_id(object_id=TEST_ID_0, object_id_info=["TEST_ID_0"])
        info_list = get_object_id_info(object_id=TEST_ID_0)
        self.assertTrue(info_list[0] == "TEST_ID_0")



import atexit
import time
import sys
from threading import Thread
from abc import abstractmethod
from collections import deque
from typing import Union

from tmtccmd.config.definitions import CoreServiceList, CoreModeList
from tmtccmd.tm.definitions import TmTypes
from tmtccmd.tm.handler import TmHandler
from tmtccmd.utility.logger import get_console_logger
from tmtccmd.sendreceive.sequential_sender_receiver import (
    SequentialCommandSenderReceiver,
)
from tmtccmd.sendreceive.tm_listener import TmListener
from tmtccmd.ccsds.handler import CcsdsTmHandler
from tmtccmd.com_if.com_interface_base import CommunicationInterface
from tmtccmd.utility.tmtc_printer import TmTcPrinter
from tmtccmd.tc.packer import ServiceQueuePacker


LOGGER = get_console_logger()


class BackendBase:
    @abstractmethod
    def initialize(self):
        """Initialize the backend. Raise RuntimeError or ValueError on failure"""

    @abstractmethod
    def start_listener(self):
        """Start the backend. Raise RuntimeError on failure"""

    @abstractmethod
    def set_mode(self, mode: int):
        """Set backend mode
        :param mode:
        :return:
        """


class TmTcHandler(BackendBase):
    """This is the primary class which handles TMTC reception. This can be seen as the backend
    in case a GUI or front-end is implemented.
    """

    def __init__(
        self,
        com_if: CommunicationInterface,
        tmtc_printer: TmTcPrinter,
        tm_listener: TmListener,
        tm_handler: TmHandler,
        init_mode: int,
        init_service: Union[str, int] = CoreServiceList.SERVICE_17.value,
        init_opcode: str = "0",
    ):
        self.mode = init_mode
        self.com_if_key = com_if.get_id()
        self.__com_if_active = False
        self.__service = init_service
        self.__op_code = init_opcode
        self.__apid = 0

        # This flag could be used later to command the TMTC Client with a front-end
        self.one_shot_operation = True

        self.__com_if = com_if
        self.__tmtc_printer = tmtc_printer
        self.__tm_listener = tm_listener
        if tm_handler.get_type() == TmTypes.CCSDS_SPACE_PACKETS:
            self.__tm_handler: CcsdsTmHandler = tm_handler
            for apid_queue_len_tuple in self.__tm_handler.get_apid_queue_len_list():
                self.__tm_listener.subscribe_ccsds_tm_handler(
                    apid_queue_len_tuple[0], apid_queue_len_tuple[1]
                )
        self.exit_on_com_if_init_failure = True
        self.single_command_package = bytearray(), None

    def get_com_if_id(self):
        return self.com_if_key

    def get_com_if(self) -> CommunicationInterface:
        return self.__com_if

    def get_printer(self) -> TmTcPrinter:
        return self.__tmtc_printer

    def get_listener(self):
        return self.__tm_listener

    def set_com_if(self, com_if: CommunicationInterface):
        if not self.is_com_if_active():
            self.__com_if = com_if
            self.__tm_listener.set_com_if(self.__com_if)
        else:
            LOGGER.warning(
                "Communication Interface is active and must be closed first before"
                "reassigning a new one"
            )

    def is_com_if_active(self):
        return self.__com_if_active

    def set_one_shot_or_loop_handling(self, enable: bool):
        """
        Specify whether the perform_operation() call will only handle one action depending
        on the mode or keep listening for replies after handling an operation.
        """
        self.one_shot_operation = enable

    def set_mode(self, mode: int):
        """
        Set the mode which will determine what perform_operation does.
        """
        self.mode = mode

    def get_mode(self) -> int:
        return self.mode

    def set_service(self, service: Union[str, int]):
        self.__service = service

    def set_opcode(self, op_code: str):
        self.__op_code = op_code

    def get_service(self) -> Union[str, int]:
        return self.__service

    def get_opcode(self) -> str:
        return self.__op_code

    def get_current_apid(self) -> int:
        return self.__apid

    def set_current_apid(self, apid: int):
        self.__apid = apid

    @staticmethod
    def prepare_tmtc_handler_start(
        com_if: CommunicationInterface,
        tmtc_printer: TmTcPrinter,
        tm_listener: TmListener,
        init_mode: int,
        init_service: Union[str, int] = CoreServiceList.SERVICE_17.value,
        init_opcode: str = "0",
    ):
        from multiprocessing import Process

        tmtc_handler = TmTcHandler(
            com_if=com_if,
            tmtc_printer=tmtc_printer,
            tm_listener=tm_listener,
            init_mode=init_mode,
            init_service=init_service,
            init_opcode=init_opcode,
        )
        tmtc_task = Process(target=TmTcHandler.start_handler, args=(tmtc_handler,))
        return tmtc_task

    @staticmethod
    def start_handler(executed_handler):
        if not isinstance(executed_handler, TmTcHandler):
            LOGGER.error("Unexpected argument, should be TmTcHandler!")
            sys.exit(1)
        executed_handler.initialize()
        executed_handler.start_listener()

    def initialize(self):
        from tmtccmd.utility.exit_handler import keyboard_interrupt_handler

        """
        Perform initialization steps which might be necessary after class construction.
        This has to be called at some point before using the class!
        """
        if self.mode == CoreModeList.LISTENER_MODE:
            LOGGER.info("Running in listener mode..")
        atexit.register(
            keyboard_interrupt_handler, tmtc_backend=self, com_interface=self.__com_if
        )

    def start_listener(self, perform_op_immediately: bool = True):
        try:
            self.__com_if.open()
            self.__tm_listener.start()
            self.__com_if_active = True
        except IOError:
            LOGGER.error("Communication Interface could not be opened!")
            LOGGER.info("TM listener will not be started")
            if self.exit_on_com_if_init_failure:
                LOGGER.error("Closing TMTC commander..")
                self.__com_if.close()
                sys.exit(1)
        if perform_op_immediately:
            self.perform_operation()

    def close_listener(self, join: bool = False, join_timeout_seconds: float = 1.0):
        """Closes the TM listener and the communication interface. This is started in a separarate
        thread because the communication interface might still be busy. The completion can be
        checked with :meth:`tmtccmd.core.backend.is_com_if_active`. Alternatively, waiting on
        completion is possible by specifying the join argument and a timeout in
        floating point second.
        :param join:
        :param join_timeout_seconds:
        :return:
        """
        if self.__com_if_active:
            close_thread = Thread(target=self.__com_if_closing)
            close_thread.start()
            if join:
                close_thread.join(timeout=join_timeout_seconds)

    def perform_operation(self):
        """Periodic operation"""
        try:
            self.__core_operation(self.one_shot_operation)
        except KeyboardInterrupt:
            LOGGER.info("Keyboard Interrupt.")
            sys.exit()
        except IOError:
            LOGGER.exception("IO Error occured")
            sys.exit()

    def __com_if_closing(self):
        self.__tm_listener.stop()
        while True:
            if not self.__tm_listener.is_listener_active():
                self.__com_if.close()
                self.__com_if_active = False
                break
            else:
                time.sleep(0.2)

    def __handle_action(self):
        """Command handling."""
        if self.mode == CoreModeList.LISTENER_MODE:
            if self.__tm_listener.reply_event():
                LOGGER.info("TmTcHandler: Packets received.")
                packet_queues = self.__tm_listener.retrieve_tm_packet_queues(clear=True)
                if len(packet_queues) > 0:
                    self.__tm_handler.handle_packet_queues(
                        packet_queue_list=packet_queues
                    )
                self.__tm_listener.clear_reply_event()
        elif self.mode == CoreModeList.SEQUENTIAL_CMD_MODE:
            service_queue = deque()
            service_queue_packer = ServiceQueuePacker()
            service_queue_packer.pack_service_queue_core(
                service=self.__service,
                service_queue=service_queue,
                op_code=self.__op_code,
            )
            if not self.__com_if.valid:
                return
            LOGGER.info("Performing service command operation")
            sender_and_receiver = SequentialCommandSenderReceiver(
                com_if=self.__com_if,
                tmtc_printer=self.__tmtc_printer,
                tm_handler=self.__tm_handler,
                tm_listener=self.__tm_listener,
                tc_queue=service_queue,
                apid=self.__apid,
            )
            sender_and_receiver.send_queue_tc_and_receive_tm_sequentially()
            self.mode = CoreModeList.LISTENER_MODE
        else:
            try:
                from tmtccmd.config.hook import get_global_hook_obj

                hook_obj = get_global_hook_obj()
                hook_obj.perform_mode_operation(mode=self.mode, tmtc_backend=self)
            except ImportError as error:
                print(error)
                LOGGER.error("Custom mode handling module not provided!")

    def __core_operation(self, one_shot):
        if self.mode == CoreModeList.LISTENER_MODE:
            one_shot = False
        if not one_shot:
            while True:
                self.__handle_action()
                if self.mode == CoreModeList.IDLE:
                    LOGGER.info("TMTC Client in idle mode")
                    time.sleep(5)
                elif self.mode == CoreModeList.LISTENER_MODE:
                    time.sleep(1)
        else:
            self.__handle_action()



"""
:file:      tcpip_udp_com_if.py
:date:      13.05.2021
:brief:     UDP Communication Interface
:author:    R. Mueller
"""
import select
import socket
from typing import Union

from tmtccmd.utility.logger import get_console_logger
from tmtccmd.com_if.com_interface_base import CommunicationInterface
from tmtccmd.tm.definitions import TelemetryListT
from tmtccmd.utility.tmtc_printer import TmTcPrinter
from tmtccmd.config.definitions import EthernetAddressT, CoreModeList

LOGGER = get_console_logger()

UDP_RECV_WIRETAPPING_ENABLED = False
UDP_SEND_WIRETAPPING_ENABLED = False


# pylint: disable=abstract-method
# pylint: disable=arguments-differ
# pylint: disable=too-many-arguments
class TcpIpUdpComIF(CommunicationInterface):
    """Communication interface for UDP communication."""

    def __init__(
        self,
        com_if_key: str,
        tm_timeout: float,
        tc_timeout_factor: float,
        send_address: EthernetAddressT,
        max_recv_size: int,
        recv_addr: Union[None, EthernetAddressT] = None,
        tmtc_printer: Union[None, TmTcPrinter] = None,
        init_mode: int = CoreModeList.LISTENER_MODE,
    ):
        """Initialize a communication interface to send and receive UDP datagrams.
        :param tm_timeout:
        :param tc_timeout_factor:
        :param send_address:
        :param max_recv_size:
        :param recv_addr:
        :param tmtc_printer:        Printer instance, can be passed optionally to allow packet debugging
        """
        super().__init__(com_if_key=com_if_key, tmtc_printer=tmtc_printer)
        self.tm_timeout = tm_timeout
        self.tc_timeout_factor = tc_timeout_factor
        self.udp_socket = None
        self.send_address = send_address
        self.recv_addr = recv_addr
        self.max_recv_size = max_recv_size
        self.init_mode = init_mode

    def __del__(self):
        try:
            self.close()
        except IOError:
            LOGGER.warning("Could not close UDP communication interface!")

    def initialize(self, args: any = None) -> any:
        pass

    def open(self, args: any = None):
        self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # Bind is possible but should not be necessary, and introduces risk of port alread
        # being used.
        # See: https://docs.microsoft.com/en-us/windows/win32/api/winsock/nf-winsock-bind
        if self.recv_addr is not None:
            LOGGER.info(
                f"Binding UDP socket to {self.recv_addr[0]} and port {self.recv_addr[1]}"
            )
            self.udp_socket.bind(self.recv_addr)
        # Set non-blocking because we use select.
        self.udp_socket.setblocking(False)
        if self.init_mode == CoreModeList.LISTENER_MODE:
            from tmtccmd.pus.service_17_test import pack_service17_ping_command

            # Send ping command immediately so the reception address is known for UDP
            ping_cmd = pack_service17_ping_command(ssc=0)
            self.send(ping_cmd.pack())

    def close(self, args: any = None) -> None:
        if self.udp_socket is not None:
            self.udp_socket.close()

    def send(self, data: bytearray):
        if self.udp_socket is None:
            return
        bytes_sent = self.udp_socket.sendto(data, self.send_address)
        if bytes_sent != len(data):
            LOGGER.warning("Not all bytes were sent!")

    def data_available(self, timeout: float = 0, parameters: any = 0) -> bool:
        if self.udp_socket is None:
            return False
        ready = select.select([self.udp_socket], [], [], timeout)
        if ready[0]:
            return True
        return False

    def receive(self, poll_timeout: float = 0) -> TelemetryListT:
        if self.udp_socket is None:
            return []
        try:
            ready = self.data_available(poll_timeout)
            if ready:
                data, sender_addr = self.udp_socket.recvfrom(self.max_recv_size)
                packet_list = [bytearray(data)]
                return packet_list
            return []
        except ConnectionResetError:
            LOGGER.warning("Connection reset exception occured!")
            return []



"""Contains core methods called by entry point files to initiate the TMTC commander.

The commander is started by first running :py:func:`initialize_tmtc_commander` and then
running :py:func:`run_tmtc_commander`
"""
import sys
import os
from typing import Union

from spacepackets.ecss.conf import get_default_tc_apid

from tmtccmd import __version__
from tmtccmd.config.hook import TmTcHookBase
from tmtccmd.core.backend import BackendBase
from tmtccmd.core.frontend_base import FrontendBase
from tmtccmd.config.definitions import CoreGlobalIds
from tmtccmd.tm.definitions import TmTypes
from tmtccmd.tm.handler import TmHandler
from tmtccmd.ccsds.handler import CcsdsTmHandler
from tmtccmd.core.globals_manager import (
    update_global,
    get_global,
    lock_global_pool,
    unlock_global_pool,
)
from tmtccmd.core.object_id_manager import insert_object_ids
from tmtccmd.config.args import parse_input_arguments
from tmtccmd.config.objects import get_core_object_ids
from tmtccmd.utility.logger import set_tmtc_console_logger, get_console_logger
from tmtccmd.utility.conf_util import AnsiColors


LOGGER = get_console_logger()


def get_tmtccmd_version() -> str:
    return __version__


def initialize_tmtc_commander(hook_object: TmTcHookBase):
    """This function needs to be called first before running the TMTC commander core. A hook
    object handle needs to be passed to this function. The user should implement an own hook class
    instance which in turn implemented TmTcHookBase. An instantiation of the hook object is then
    passed to the core. The hook object ecncapsulates the control of the user over the TMTC
    commander core.

    Example for a simple main function content to use the command line mode:

        hook_obj = MyCustomHookClass()
        initialize_tmtccmd(hook_obj)
        run_tmtc_client(False)

    :param hook_object:     Instantiation of a custom hook object. The TMTC core will call the
                            various hook functions during program run-time.
    :raises: ValueError for an invalid hook object.
    """
    if os.name == "nt":
        import colorama

        colorama.init()

    __assign_tmtc_commander_hooks(hook_object=hook_object)


def add_ccsds_handler(ccsds_handler: CcsdsTmHandler):
    """Add a handler for CCSDS space packets, for example PUS packets
    :param pus_handler:
    :return:
    """
    lock_global_pool()
    tm_handler = get_global(CoreGlobalIds.TM_HANDLER_HANDLE)
    if tm_handler is None:
        update_global(CoreGlobalIds.TM_HANDLER_HANDLE, ccsds_handler)
    unlock_global_pool()


def run_tmtc_commander(
    use_gui: bool,
    reduced_printout: bool = False,
    ansi_colors: bool = True,
    tmtc_backend: Union[BackendBase, None] = None,
    tmtc_frontend: Union[FrontendBase, None] = None,
    app_name: str = "TMTC Commander",
):
    """This is the primary function to run the TMTC commander. Users should call this function to
    start the TMTC commander. Please note that :py:func:`initialize_tmtc_commander` needs to be
    called before this function. Raises RuntimeError if :py:func:`initialize_tmtc_commander`
    has not been called before calling this function.

    :param use_gui:             Specify whether the GUI is used or not
    :param reduced_printout:    It is possible to reduce the initial printout with this flag
    :param ansi_colors:         Enable ANSI color output for terminal
    :param tmtc_backend:        Custom backend can be passed here. Otherwise, a default backend
                                will be created
    :param tmtc_frontend:       Custom frontend can be passed here. Otherwise, a default backend
                                will be created
    :param app_name:            Name of application. Will be displayed in GUI
    :raises RunTimeError:  if :py:func:`initialize_tmtc_commander` was not called before
    :return:
    """
    try:
        __set_up_tmtc_commander(
            use_gui=use_gui, reduced_printout=reduced_printout, ansi_colors=ansi_colors
        )
    except ValueError:
        raise RuntimeError

    if use_gui:
        __start_tmtc_commander_qt_gui(tmtc_frontend=tmtc_frontend, app_name=app_name)
    else:
        if tmtc_backend is None:
            from tmtccmd.config.hook import get_global_hook_obj

            hook_obj = get_global_hook_obj()
            json_cfg_path = hook_obj.get_json_config_file_path()
            tm_handler = get_global(CoreGlobalIds.TM_HANDLER_HANDLE)
            tmtc_backend = get_default_tmtc_backend(
                hook_obj=hook_obj, tm_handler=tm_handler, json_cfg_path=json_cfg_path
            )
        __start_tmtc_commander_cli(tmtc_backend=tmtc_backend)


def __assign_tmtc_commander_hooks(hook_object: TmTcHookBase):
    if hook_object is None:
        raise ValueError

    # Check whether all required hook functions have bee implemented properly, Python
    # does not enforce this.
    if (
        hook_object.add_globals_pre_args_parsing is None
        or hook_object.add_globals_post_args_parsing is None
    ):
        LOGGER.error(
            "Passed hook base object handle is invalid. Abstract functions have to be implemented!"
        )
        raise ValueError
    # Insert hook object handle into global dictionary so it can be used by the TMTC commander
    update_global(CoreGlobalIds.TMTC_HOOK, hook_object)
    # Set core object IDs
    insert_object_ids(get_core_object_ids())
    # Set object IDs specified by the user.
    insert_object_ids(hook_object.get_object_ids())


def __set_up_tmtc_commander(
    use_gui: bool,
    reduced_printout: bool,
    ansi_colors: bool = True,
    tmtc_backend: Union[BackendBase, None] = None,
):
    """Set up the TMTC commander. Raise ValueError if a passed parameter is invalid.
    :param use_gui:
    :param reduced_printout:
    :param ansi_colors:
    :param tmtc_backend:
    :return:
    """
    from tmtccmd.config.hook import TmTcHookBase
    from typing import cast

    set_tmtc_console_logger()

    # First, we check whether a hook object was passed to the TMTC commander. This hook object
    # encapsulates control of the commnader core so it is required for proper functioning
    # of the commander core.
    hook_obj_raw = get_global(CoreGlobalIds.TMTC_HOOK)
    if hook_obj_raw is None:
        LOGGER.warning(
            "No valid hook object found. initialize_tmtc_commander needs to be called first. "
            "Terminating.."
        )
        raise ValueError
    hook_obj = cast(TmTcHookBase, hook_obj_raw)

    if not reduced_printout:
        __handle_init_printout(use_gui, ansi_colors)

    LOGGER.info("Starting TMTC Commander..")

    if use_gui:
        hook_obj.add_globals_pre_args_parsing(True)
    else:
        __handle_cli_args_and_globals()


def __handle_init_printout(use_gui: bool, ansi_colors: bool):
    if ansi_colors:
        print(f"{AnsiColors.CYAN}-- Python TMTC Commander --{AnsiColors.RESET}")
    if use_gui:
        print("-- GUI mode --")
    else:
        print("-- Command line mode --")

    print(f"-- tmtccmd version v{get_tmtccmd_version()} --")


def __handle_cli_args_and_globals():
    from typing import cast
    from tmtccmd.core.globals_manager import get_global

    hook_obj = cast(TmTcHookBase, get_global(CoreGlobalIds.TMTC_HOOK))
    LOGGER.info("Setting up pre-globals..")
    hook_obj.add_globals_pre_args_parsing(False)

    LOGGER.info("Parsing input arguments..")
    args = parse_input_arguments()
    LOGGER.info("Setting up post-globals..")
    hook_obj.add_globals_post_args_parsing(args=args)


def __start_tmtc_commander_cli(tmtc_backend: BackendBase):
    __get_backend_init_variables()
    tmtc_backend.initialize()
    tmtc_backend.start_listener()


def __start_tmtc_commander_qt_gui(
    tmtc_frontend: Union[None, FrontendBase] = None, app_name: str = "TMTC Commander"
):
    app = None
    if tmtc_frontend is None:
        from tmtccmd.core.frontend import TmTcFrontend
        from tmtccmd.config.hook import get_global_hook_obj

        try:
            from PyQt5.QtWidgets import QApplication
        except ImportError:
            LOGGER.error("PyQt5 module not installed, can't run GUI mode!")
            sys.exit(1)
        app = QApplication([app_name])
        hook_obj = get_global_hook_obj()
        json_cfg_path = hook_obj.get_json_config_file_path()
        tm_handler = get_global(CoreGlobalIds.TM_HANDLER_HANDLE)
        # The global variables are set by the argument parser.
        tmtc_backend = get_default_tmtc_backend(
            hook_obj=hook_obj, tm_handler=tm_handler, json_cfg_path=json_cfg_path
        )
        tmtc_frontend = TmTcFrontend(
            hook_obj=hook_obj, tmtc_backend=tmtc_backend, app_name=app_name
        )
    tmtc_frontend.start(app)


def __get_backend_init_variables():
    service = get_global(CoreGlobalIds.CURRENT_SERVICE)
    op_code = get_global(CoreGlobalIds.OP_CODE)
    com_if = get_global(CoreGlobalIds.COM_IF)
    mode = get_global(CoreGlobalIds.MODE)
    return service, op_code, com_if, mode


def get_default_tmtc_backend(
    hook_obj: TmTcHookBase, tm_handler: TmHandler, json_cfg_path: str
):
    from tmtccmd.core.backend import TmTcHandler
    from tmtccmd.utility.tmtc_printer import TmTcPrinter
    from tmtccmd.sendreceive.tm_listener import TmListener
    from typing import cast

    service, op_code, com_if_id, mode = __get_backend_init_variables()
    display_mode = get_global(CoreGlobalIds.DISPLAY_MODE)
    print_to_file = get_global(CoreGlobalIds.PRINT_TO_FILE)
    tmtc_printer = TmTcPrinter(display_mode, print_to_file, True)
    if tm_handler is None:
        LOGGER.warning("No TM Handler specified!")
    else:
        if tm_handler.get_type() == TmTypes.CCSDS_SPACE_PACKETS:
            tm_handler = cast(CcsdsTmHandler, tm_handler)
            tm_handler.initialize(tmtc_printer=tmtc_printer)
    apid = get_default_tc_apid()

    if json_cfg_path:
        pass
    com_if = hook_obj.assign_communication_interface(
        com_if_key=get_global(CoreGlobalIds.COM_IF), tmtc_printer=tmtc_printer
    )
    tc_send_timeout_factor = get_global(CoreGlobalIds.TC_SEND_TIMEOUT_FACTOR)
    tm_timeout = get_global(CoreGlobalIds.TM_TIMEOUT)
    tm_listener = TmListener(
        com_if=com_if, tm_timeout=tm_timeout, tc_timeout_factor=tc_send_timeout_factor
    )
    # The global variables are set by the argument parser.
    tmtc_backend = TmTcHandler(
        com_if=com_if,
        tmtc_printer=tmtc_printer,
        tm_listener=tm_listener,
        init_mode=mode,
        init_service=service,
        init_opcode=op_code,
        tm_handler=tm_handler,
    )
    tmtc_backend.set_current_apid(apid=apid)
    tmtc_backend.set_one_shot_or_loop_handling(
        get_global(CoreGlobalIds.USE_LISTENER_AFTER_OP)
    )
    return tmtc_backend



from __future__ import annotations
from typing import Union
import struct
from tmtccmd.utility.logger import get_console_logger

LOGGER = get_console_logger()


class ObjectId:
    def __init__(self, object_id: int):
        self.id = object_id

    @classmethod
    def from_bytes(cls, obj_id_as_bytes: bytearray) -> ObjectId:
        obj_id = ObjectId(object_id=0)
        obj_id.id = obj_id_as_bytes
        return obj_id

    @property
    def id(self) -> int:
        return self._object_id

    @id.setter
    def id(self, new_id: Union[int, bytearray]):
        if isinstance(new_id, int):
            self._object_id = new_id
            self._id_as_bytes = struct.pack("!I", self._object_id)
        elif isinstance(new_id, bytearray):
            if len(new_id) != 4:
                LOGGER.warning(f"Invalid object ID length {len(new_id)}")
                raise ValueError
            self._id_as_bytes = new_id
            self._object_id = struct.unpack("!I", self._id_as_bytes[:])[0]
        else:
            raise ValueError

    @property
    def as_int(self) -> int:
        return self._object_id

    @property
    def as_bytes(self) -> bytes:
        return self._id_as_bytes

    @property
    def as_string(self) -> str:
        return f"0x{self._object_id:08x}"



from tmtccmd.tm.service_3_housekeeping import Service3TM

from tmtccmd.tm.definitions import TelemetryListT, TelemetryQueueT
from tmtccmd.tm.service_5_event import Service5TM
from tmtccmd.tm.service_8_functional_cmd import Service8TM



from unittest import TestCase
from tmtccmd.config.globals import (
    update_global,
    get_global,
    set_json_cfg_path,
    get_json_cfg_path,
    set_glob_com_if_dict,
    get_glob_com_if_dict,
    set_default_globals_pre_args_parsing,
    check_and_set_core_mode_arg,
    CoreModeList,
    CoreGlobalIds,
)


class TestGlobalManager(TestCase):
    def test_global_module(self):
        update_global(global_param_id=30, parameter="hello")
        self.assertTrue(get_global(global_param_id=30) == "hello")

        current_path = get_json_cfg_path()
        set_json_cfg_path(".")
        self.assertTrue(get_json_cfg_path() == ".")
        set_json_cfg_path(current_path)

        custom_com_if_dict = {"test": ("Test Interface", "")}
        set_glob_com_if_dict(custom_com_if_dict=custom_com_if_dict)
        com_if_dict = get_glob_com_if_dict()
        self.assertTrue(com_if_dict["test"][0] == "Test Interface")

        set_default_globals_pre_args_parsing(gui=False, tc_apid=0x02, tm_apid=0x03)

        result = check_and_set_core_mode_arg(mode_arg="udp")
        self.assertTrue(result == CoreModeList.SEQUENTIAL_CMD_MODE)

        result = check_and_set_core_mode_arg(mode_arg="listener")
        self.assertTrue(get_global(CoreGlobalIds.MODE) == CoreModeList.LISTENER_MODE)
        self.assertTrue(result == CoreModeList.LISTENER_MODE)

        result = check_and_set_core_mode_arg(mode_arg="seqcmd")
        self.assertTrue(
            get_global(CoreGlobalIds.MODE) == CoreModeList.SEQUENTIAL_CMD_MODE
        )
        self.assertTrue(result == CoreModeList.SEQUENTIAL_CMD_MODE)
